/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import gnu.getopt.*;
import java.io.*;
import netscape.ldap.*;
import org.w3c.dom.Document;

/** 

  This class exposes functionality to import a DSML file into a directory.
  <BR><BR>
  This program reads in DSML, either from a file or from stdin, connects to
  a specified LDAP server and executes add or delete operations based on
  options set and the contents of the DSML file.<BR>
  <BR><PRE>
  The command syntax is as follows. Bold indicates a mandatory option.<BR>
  DSML2LDAP infile <B>-h ldap.example.com [-a|-d]</B> [options]<BR>
  Connection options:<UL>
  <LI>--host      -h LDAP host.
  <LI>--port      -p Port on the server to connect to. Default - 389.
  <LI>--username  -u Username. Omit for anonymous login.
  <LI>--password  -w Password. Omit for anonymous login.
  </UL>
  Configuration options:<UL>
  <LI>--add       -a Adds the contents of the DSML file to the directory
  <LI>--delete    -d Deletes the contents of the DSML file from the
  directory         
  You must specify one of -a or -d.
  <LI>--careful   -c In "add" mode, this means we don't overwrite entries
  that are already present. In "delete" mode, this means 
  we only delete entries that are an exact match, rather
  than just matching on DN.         
  </UL>
  Other options:<UL>
  <LI>--quick     -q Quick mode - do not validate DSML documents
  <LI>--help      -? Prints a usage message.
  </UL>
  E.g.:
  DSML2LDAP in.dsml -h ldap.ex.com -u "cn=Fred Bloggs" -w SecretWord -a -c
  </PRE>
  */
public class DSML2LDAP extends DSMLFile
{
/** 

  Stores input file name
  */
  private static String  s_infileName = null;

/** 

  Stores hostname to connect to. Mandatory
  */
  private static String  s_hostname   = null;           
  
/** 

  Stores port to connect on
  */
  private static int     s_port       = LDAPv2.DEFAULT_PORT;
  
/** 

  Stores username to connect with (entry in directory)
  */
  private static String  s_username   = null;
  
/** 

  Stores password to connect with
  */
  private static String  s_password   = null;
  
/** 

  Add contents of DSML to directory? (False == delete it)
  */
  private static boolean s_add;      /* -a or -d is mandatory, so no default */
  
/** 

  In the case of add, true here means "don't add already-present
  entries"; in the case of delete, true means "don't overwrite existing
  entries".
  */
  private static boolean s_careful    = false;
  
/** 

  Avoid doing directory-style validation of the DSML?
  */
  private static boolean s_quick      = false;

/** This method drives the DSML2LDAP process.

  @param xiArgs The command-line options
  */
  public static void main(String[] xiArgs)
  {
    Document lDoc = null;

    /*************************************************************************/
    /* Parse command line                                                    */
    /*************************************************************************/
    parseCommandLine(xiArgs);
    
    /*************************************************************************/
    /* Read in the DSML; if s_infileName is null, it will read from stdin.   */
    /*************************************************************************/
    lDoc = readDSML(s_infileName);

    /*************************************************************************/
    /* Check for "complete" attribute; if not complete, don't validate.      */
    /*************************************************************************/
    if (((lDoc.getDocumentElement()).getAttribute("complete")).equals("false"))
    {
      s_quick = true;
      System.err.print("Unable to validate document");

      if (s_infileName != null)
      {
        System.err.print(" '" + s_infileName + "'");
      }

      System.err.println(" as it contains references to an external schema.");
    }

    /*************************************************************************/
    /* Validate if required                                                  */
    /*************************************************************************/
    if (!s_quick)
    {
      try
      {
        System.err.print("Validating document");

        if (s_infileName != null)
        {
          System.err.print(" '" + s_infileName + "'");
        }

        System.err.println(".");
        DSMLDocumentImpl.validate(lDoc);
      }
      catch (DSMLRefException lExcept)
      {
        System.err.println("Unable to validate document \"" + s_infileName + 
                        "\" as it contains references to an external schema.");
      }
    }

    /*************************************************************************/
    /* Create a DSMLLDAPWriter.  It does its own error handling.             */
    /*************************************************************************/
    DSMLLDAPWriter lWriter = new DSMLLDAPWriter();

    /*************************************************************************/
    /* Configure the DSMLLDAPWriter                                          */
    /*************************************************************************/
    lWriter.setCareful(s_careful);
    lWriter.connect(s_hostname, s_port, s_username, s_password);

    /*************************************************************************/
    /* Call the correct method (add or delete)                               */
    /*************************************************************************/
    if (s_add)
    {
      lWriter.add(lDoc);
    }
    else
    {
      lWriter.delete(lDoc);
    }

    lWriter.disconnect();

  } /* main() */

/** This function creates a GetOpt object, passing in the argument string,
    and makes repeated calls to getopt() to set all or some of the
    command-line option member variables.

  @param xiArgs The argument string to parse
  */
  private static void parseCommandLine(String[] xiArgs)
  {
    boolean lAddDefined = false;
    boolean lDelDefined = false;
    int     lOpt        = 0;

    LongOpt[] lLOs = new LongOpt[9];
    lLOs[0] = new LongOpt("hostname", LongOpt.REQUIRED_ARGUMENT, null, 'h');
    lLOs[1] = new LongOpt("port",     LongOpt.REQUIRED_ARGUMENT, null, 'p');
    lLOs[2] = new LongOpt("username", LongOpt.REQUIRED_ARGUMENT, null, 'u');
    lLOs[3] = new LongOpt("password", LongOpt.REQUIRED_ARGUMENT, null, 'w');
    lLOs[4] = new LongOpt("add",      LongOpt.NO_ARGUMENT, null, 'a');
    lLOs[5] = new LongOpt("delete",   LongOpt.NO_ARGUMENT, null, 'd');
    lLOs[6] = new LongOpt("careful",  LongOpt.NO_ARGUMENT, null, 'c');
    lLOs[7] = new LongOpt("quick",    LongOpt.NO_ARGUMENT, null, 'q');
    lLOs[8] = new LongOpt("help",     LongOpt.NO_ARGUMENT, null, '?');

    Getopt lGetopt = new Getopt("DSML2LDAP", 
                                xiArgs, 
                                "h:p:u:w:adcq?",
                                lLOs);

    /*************************************************************************/
    /* Keep parsing until we run out of options                              */
    /*************************************************************************/
    while ((lOpt = lGetopt.getopt()) != -1)
    {
      switch (lOpt)
      {
        /*********************************************************************/
        /* Hostname                                                          */
        /*********************************************************************/
        case 'h':
          s_hostname = lGetopt.getOptarg();
          break;
      
        /*********************************************************************/
        /* Port                                                              */
        /*********************************************************************/
        case 'p':
          String lPort = lGetopt.getOptarg();
          try
          {
            s_port = Integer.parseInt(lPort);
          }
          catch (NumberFormatException lExcept)
          {
            System.err.println("DSML2LDAP: Invalid argument to -p: " + lPort +
                               ". Not a number.");
            printUsageMessage();
            System.exit(1);
          }
          break;

        /*********************************************************************/
        /* Username                                                          */
        /*********************************************************************/
        case 'u':
          s_username = lGetopt.getOptarg();
          break;

        /*********************************************************************/
        /* Password                                                          */
        /*********************************************************************/
        case 'w':
          s_password = lGetopt.getOptarg();
          break;

        /*********************************************************************/
        /* Adding?                                                           */
        /*********************************************************************/
        case 'a':
          if (!lDelDefined)
          {
            s_add = true;
            lAddDefined = true;
          }
          else
          {
            System.err.println("DSML2LDAP: Both of -a and -d specified. " +
                               "Please specify one only.");
            printUsageMessage();
            System.exit(1);
          }
          break;

        /*********************************************************************/
        /* Deleting?                                                         */
        /*********************************************************************/
        case 'd':
          if (!lAddDefined)
          {
            s_add = false;
            lDelDefined = true;
          }
          else
          {
            System.err.println("DSML2LDAP: Both of -a and -d specified. " +
                               "Please specify one only.");
            printUsageMessage();
            System.exit(1);
          }
          break;

        /*********************************************************************/
        /* Careful?                                                          */
        /*********************************************************************/
        case 'c':
          s_careful = true;
          break;

        /*********************************************************************/
        /* Turn off validation?                                              */
        /*********************************************************************/
        case 'q':
          s_quick = true;
          break;

        /*********************************************************************/
        /* Print usage message?                                              */
        /*                                                                   */
        /* This value (?) can be requested on the command line, but is also  */
        /* returned when GetOpt encounters an error.                         */
        /*********************************************************************/
        case '?':
          printUsageMessage();
          System.exit(0);
          break;
        
        /*********************************************************************/
        /* Default - invalid option                                          */
        /*********************************************************************/
        default:
          System.err.println("DSML2LDAP: Invalid option: " + lOpt);
          printUsageMessage();
          System.exit(1);
          break;
      }
    }

    /*************************************************************************/
    /* Get non-option command line parameter (infilename).  There should     */
    /* only be one of these.  Optind is left pointing at it.                 */
    /*************************************************************************/
    int lIndex = lGetopt.getOptind();

    try
    {
      s_infileName = xiArgs[lIndex++];
    }
    catch (ArrayIndexOutOfBoundsException lExcept)
    {
      /***********************************************************************/
      /* If there are no non-option arguments, getopt sets Optind to one     */
      /* past the end of the array (not shockingly helpful.) So, this        */
      /* exception means that no infileName was specified; we'll use stdin.  */
      /***********************************************************************/
      s_infileName = null;
    }

    /*************************************************************************/
    /* If we haven't fallen off the end of xiArgs yet, there are too many    */
    /* command-line parameters.                                              */
    /*************************************************************************/
    if (lIndex < xiArgs.length)
    {
      System.err.println("DSML2LDAP: Too many filenames specified.");
      printUsageMessage();
      System.exit(1);
    }

    /*************************************************************************/
    /* Check for mandatory attributes                                        */
    /*************************************************************************/
    if (s_hostname == null)
    {
      System.err.println("DSML2LDAP: You must supply a hostname.");
      printUsageMessage();
      System.exit(1);
    }

    /*************************************************************************/
    /* Set s_add                                                             */
    /*************************************************************************/
    if      (lAddDefined)
    {
      s_add = true;
    }
    else if (lDelDefined)
    {
      s_add = false;
    }
    else /* neither -a or -d defined */
    {
      System.err.println("DSML2LDAP: You must specify either -a or -d.");
      printUsageMessage();
      System.exit(1);
    }

  } /* parseCommandLine() */

  /***************************************************************************/
  /* Method: printUsageMessage()                                             */
  /*                                                                         */
  /* This method prints a usage message - it should be called if the user    */
  /* incorrectly specifies one or more command-line parameters.              */
  /***************************************************************************/
  private static void printUsageMessage()
  {
    System.err.print(
    "\n" +
    "This command reads information from a DSML file and writes it to an LDAP\n" +
    "directory.\n" +
    "\n" +
    "The command syntax is as follows (with commands in standard Unix format):\n" +
    "DSML2LDAP <infile> -h <ldap.server.name> [-a|-d] [options]\n" +
    "\n" +
    "Connection options:\n" +
    "--host       -h LDAP host - takes a DNS name. Mandatory.\n" +
    "--port       -p Port on the server to connect to. Default - 389.\n" +
    "--username   -u Username. Omit for anonymous login.\n" +
    "--password   -w Password. Omit for anonymous login.\n" +
    "Configuration options:\n" +
    "--add        -a Adds the contents of the DSML file to the directory\n" +   
    "--delete     -d Deletes the contents of the DSML file from the directory\n" +
    "    [ You must specify one of -a or -d. ]\n" +
    "--careful    -c In \"add\" mode, this means we don't overwrite entries that\n" + 
    "                are already present. In \"delete\" mode, this means we only\n" + 
    "                delete entries that are an exact match, rather than just\n" + 
    "                matching on DN.\n" +
    "--quick      -q Quick mode - do not validate DSML documents\n" +
    "--help       -? Prints this usage message.\n" +
    "\n"
    );
  }

  /***************************************************************************/
  /* TEST_CODE                                                               */
  /***************************************************************************/
  private static void printMemberVariables()
  {  
    System.err.print("s_infileName: ");
    System.err.println(s_infileName);
    
    System.err.print("s_hostname: ");
    System.err.println(s_hostname);
    System.err.print("s_port: ");
    System.err.println(s_port);
    System.err.print("s_username: ");
    System.err.println(s_username);
    System.err.print("s_password: ");
    System.err.println(s_password);
    System.err.print("s_add: ");
    System.err.println(s_add);
    System.err.print("s_careful: ");
    System.err.println(s_careful);
    System.err.print("s_quick: ");
    System.err.println(s_quick);
  }

} /* DSML2LDAP */
