/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.util.*;
import java.io.*;
import org.apache.xml.serialize.*; 
import org.w3c.dom.*;
import org.apache.xerces.dom.DocumentImpl;

/** 

  This class implements asymmetric diffing of two DSML Documents, with the
  results returned as two more DSML Documents.
  */
public class DSMLDOMDiffer
{
/** 

  The first document.  These two documents are member variables because
  they are needed throughout this class for their factory methods to
  create Nodes and Elements.
  */
  private Document m_Doc1        = null;

/** 

  The second document
  */
  private Document m_Doc2        = null;

/** 

  Whether to add human-readable comments to the Add Document
  */
  private boolean  m_addComments = true;

/** This method creates, in Document form, the diffs necessary to turn
    xiDoc1 into xiDoc2. These diffs will be returned using the variables
    xoDocSub, and xoDocAdd, which are expected to be passed in as
    bare-bones DSML Documents.
    <B>Warning: xiDoc1 and xiDoc2 get munged by this process.</B> Don't
    pass Documents in you want to keep.

  @param xiDoc1 Base document for the diff
  @param xiDoc2 Document to be compared to xiDoc1
  @param xoDocSub Newly created DSMLDocumentImpl to hold entries to be
    subtracted.
  @param xoDocAdd Newly created DSMLDocumentImpl to hold entries to be
    added.
  */
  public void diff(Document     xiDoc1,
                   Document     xiDoc2,
                   DocumentImpl xoDocSub, 
                   DocumentImpl xoDocAdd)
  {
    /*************************************************************************/
    /* We make these member variables because we need their factory methods  */
    /* to create Nodes and Elements.                                         */
    /*************************************************************************/
    m_Doc1 = xiDoc1;
    m_Doc2 = xiDoc2;

    /*************************************************************************/
    /* Local variables                                                       */
    /*************************************************************************/
    Node    lDSMLNode   = null;              /* <dsml> element of a Document */
    Node    lNode       = null;                               /* any old node */
    Element lEntryElem1 = null;       /* an <entry> Element relating to Doc1 */
    Element lEntryElem2 = null;       /* an <entry> Element relating to Doc2 */
    Element lAttrElem1  = null;        /* an <attr> Element relating to Doc1 */
    Element lAttrElem2  = null;        /* an <attr> Element relating to Doc1 */

    Hashtable lMatchingEntries = new Hashtable();  
                                        /* Those <entry>s which are the same */
    HashSet   lRemoveMe = new HashSet();  
                             /* <entry>s to be removed from lMatchingEntries */
    
    Iterator lIterator = null;                 /* a general-purpose Iterator */
    int      lCount    = 0;
                                  
    Element  lEntriesElem1;     /* the <directory-entries> Element for Doc 1 */
    Element  lEntriesElem2;     /* the <directory-entries> Element for Doc 2 */

    NodeList lEntriesElems;         
    lEntriesElems = m_Doc1.getElementsByTagName("directory-entries");
    lEntriesElem1 = (Element)lEntriesElems.item(0);
    lEntriesElems = m_Doc2.getElementsByTagName("directory-entries");
    lEntriesElem2 = (Element)lEntriesElems.item(0);

    /*************************************************************************/
    /* Diff the documents at <entry> level; read the comment for             */
    /* diffElements to see exactly what this call does.                      */
    /*************************************************************************/
    diffElements(lEntriesElem1, 
                 lEntriesElem2, 
                 "entry", 
                 "dn",
                 2,
                 lMatchingEntries);

    /*************************************************************************/
    /* Iterate over the matching entries                                     */
    /*************************************************************************/
    Enumeration lEntriesEnum = lMatchingEntries.keys();

    while (lEntriesEnum.hasMoreElements())
    {
      lEntryElem1 = (Element)lEntriesEnum.nextElement();
      lEntryElem2 = (Element)lMatchingEntries.get(lEntryElem1);

      /***********************************************************************/
      /* Currently, lMatchingEntries contains all those entries which match  */
      /* by DN. If we remove those from xiDoc1, the remainder will be        */
      /* exactly those that need to go into xoDocSub.                        */
      /*                                                                     */
      /* It is safe to "remove" this Node from the tree - it doesn't get     */
      /* destroyed until we remove all references to it. We don't want to    */
      /* try adding more nodes to it, though - but that's OK, as we don't    */
      /* try and do that.                                                    */
      /***********************************************************************/
      String lDN = lEntryElem1.getAttribute("dn");
      lEntriesElem1.removeChild(lEntryElem1);         
      
      /***********************************************************************/
      /* Diff individual entries at <oc-value> level; mark for removal if    */
      /* there's a difference.                                               */
      /***********************************************************************/
      if (diffElements(lEntryElem1, 
                       lEntryElem2, 
                       "oc-value", 
                       null, 
                       4,
                       null))
      {
        lRemoveMe.add(lEntryElem1);
      }

      Hashtable lMatchingAttrs = new Hashtable();

      /***********************************************************************/
      /* Diff individual entries at <attr> level; mark for removal if        */
      /* there's a difference.                                               */
      /***********************************************************************/
      if (diffElements(lEntryElem1, 
                       lEntryElem2, 
                       "attr", 
                       "name", 
                       3,
                       lMatchingAttrs))
      {
        lRemoveMe.add(lEntryElem1);
      }

      Enumeration lAttrEnum = lMatchingAttrs.keys();

      /***********************************************************************/
      /* Iterate over the matching <attr>s                                   */
      /***********************************************************************/
      while (lAttrEnum.hasMoreElements())
      {
        lAttrElem1 = (Element)lAttrEnum.nextElement();
        lAttrElem2 = (Element)lMatchingAttrs.get(lAttrElem1);
        
        /*********************************************************************/
        /* Diff individual attrs at <value> level; mark for removal if       */
        /* there's a difference.                                             */
        /*********************************************************************/
        if (diffElements(lAttrElem1, 
                         lAttrElem2, 
                         "value", 
                         null, 
                         4,
                         null))
        {
          lRemoveMe.add(lEntryElem1);
        }
      }
    }

    /*************************************************************************/
    /* We aren't allowed to modify lMatchingEntries while we are Enumerating */
    /* it, so, we stored those keys to remove in lRemoveMe.  We can now      */
    /* remove them.                                                          */
    /*************************************************************************/
    lIterator = lRemoveMe.iterator();

    while (lIterator.hasNext())
    {
      lMatchingEntries.remove((Element)lIterator.next());
    }

    
    /*************************************************************************/
    /* lMatchingEntries now contains only those entries which match exactly. */
    /* If we remove these entries from the m_Doc2, the entries which make up */
    /* xoDocAdd will remain.  When we remove the "values" from m_Doc2, we    */
    /* extract them as a Collection first, as Hashtable.containsValue() is   */
    /* an expensive operation.                                               */
    /*************************************************************************/
    Collection lValues = lMatchingEntries.values();
    
    NodeList lEntries2 = m_Doc2.getElementsByTagName("entry");
    
    /*************************************************************************/
    /* Iterate over all the &lt;entry&gt;s in m_Doc2 and remove them if they */
    /* are on the "exact match" list.                                        */
    /*************************************************************************/
    for (lCount = 0; lCount < lEntries2.getLength(); lCount++)
    {
      lEntryElem2 = (Element)lEntries2.item(lCount);

      if (lValues.contains(lEntryElem2))
      {
        lEntriesElem2.removeChild(lEntryElem2);
        lCount--;
      }
    }
    
    /*************************************************************************/
    /* Now, the &lt;entry&gt;s in m_Doc1 are those needed for xoDocSub, and  */
    /* those in m_Doc2 are those needed for xoDocAdd.  So, we just adopt     */
    /* their &lt;directory-entries&gt; nodes.  This approach has the         */
    /* advantage that it also adopts any comments we added to m_Doc2 at the  */
    /* same level as the &lt;entry&gt; nodes.                                */
    /*************************************************************************/
    lDSMLNode = xoDocSub.getDocumentElement();
    lNode = xoDocSub.importNode(lEntriesElem1, true);
    lDSMLNode.appendChild(lNode);

    /*************************************************************************/
    /* xoDocSub has no schema, and therefore is not "complete".              */
    /*************************************************************************/
    ((Element)lDSMLNode).setAttribute("complete", "false");

    lDSMLNode = xoDocAdd.getDocumentElement();
    lNode = xoDocAdd.importNode(lEntriesElem2, true);
    lDSMLNode.appendChild(lNode);

    /*************************************************************************/
    /* Create mini schema for xoDocAdd from m_Doc2.                          */
    /*************************************************************************/
    createMiniSchema(xoDocAdd, m_Doc2);

  } /* diff() */

/** Creates a mini schema.  The other implementation of createMiniSchema
    does it in terms of Netscape LDAP structures.  Doing the conversion
    both ways here would be horribly inefficient, so we have to reimplement
    the functionality in Document terms.  <sigh>

  @param xbEntryDoc The document containing the entries we are producing a
    schema for, and to which we will add the mini schema.
  @param xiSchemaDoc The document containing the full schema from which we will
    be extracting bits.
  */
  private static void createMiniSchema(DocumentImpl xbEntryDoc,
                                       Document     xiSchemaDoc)
  {
    /*************************************************************************/
    /* Local variables                                                       */
    /*************************************************************************/
    int  lLength = 0;
    int  lCount  = 0;
    Node lNode   = null;

    /*************************************************************************/
    /* Create a &lt;directory-schema&gt; node in xbEntryDoc, as the first    */
    /* child node of &lt;dsml&gt;.                                           */
    /*************************************************************************/
    
    Element lDSMLElem   = (Element)xbEntryDoc.getDocumentElement();
    Element lSchemaElem = xbEntryDoc.createElement("directory-schema");
    lDSMLElem.insertBefore(lSchemaElem, lDSMLElem.getFirstChild());
    
    /*************************************************************************/
    /* Get all the objectclass names from the <entry>s (so we know what      */
    /* objectclasses are used by the data) and add them to a hash.  The      */
    /* objectclass names are all the text children of <oc-value> nodes.      */
    /*************************************************************************/
    NodeList lOCValueElems  = xbEntryDoc.getElementsByTagName("oc-value");
    Element  lOCValueElem   = null;
    String   lOCValue       = null;

    lLength = lOCValueElems.getLength();
    
    HashSet lOCNames = new HashSet(lLength);

    for (lCount = 0; lCount < lLength; lCount++)
    {
      lOCValueElem = (Element)lOCValueElems.item(lCount);
      lNode = lOCValueElem.getFirstChild();
      
      /***********************************************************************/
      /* Skip past any comments there might be                               */
      /***********************************************************************/
      while (lNode.getNodeType() == Node.COMMENT_NODE)
      {
        lNode = lNode.getNextSibling();
      }

      lOCValue = lNode.getNodeValue();
      lOCNames.add(lOCValue);
    }

    /*************************************************************************/
    /* There are two sorts of <name> tag in a DSML file those which denote   */
    /* attribute names and those which denote objectclass names (both in the */
    /* schema).                                                              */
    /*************************************************************************/
    NodeList lNameElems = xiSchemaDoc.getElementsByTagName("name");
    Element  lNameElem  = null;
    String   lName      = null;

    for (lCount = 0; lCount < lNameElems.getLength(); lCount++)
    {
      lNameElem = (Element)lNameElems.item(lCount);
      lNode = lNameElem.getFirstChild();
      
      /***********************************************************************/
      /* Skip past any comments there might be                               */
      /***********************************************************************/
      while (lNode.getNodeType() == Node.COMMENT_NODE)
      {
        lNode = lNode.getNextSibling();
      }

      lName = lNode.getNodeValue();

      /***********************************************************************/
      /* You tell which sort of name node you have by looking at the node    */
      /* name of the parent node.  It will be either "class" or "attr".      */
      /***********************************************************************/
      if (((lNameElem.getParentNode()).getNodeName()).compareTo("class") == 0)
      {
        /*********************************************************************/
        /* It's an objectclass name.  If we are using this objectclass, we   */
        /* need to add the <class> Element to the mini schema.               */
        /*********************************************************************/
        if (lOCNames.contains(lName))
        {
          lNode = lNameElem.getParentNode();
          lNode = xbEntryDoc.importNode(lNode, true);
          lSchemaElem.appendChild(lNode); 
        }
        else
        {
        }
      }
    }

    /*************************************************************************/
    /* Now, the mini schema has all the required objectclasses. Next, we see */
    /* what attribute-types those objectclasses use.                         */
    /*************************************************************************/
    NodeList lAttrElems = xbEntryDoc.getElementsByTagName("attribute");
    Element  lAttrElem  = null;
    String   lAttr      = null;

    lLength = lAttrElems.getLength();
    
    HashSet lAttrNames = new HashSet();

    for (lCount = 0; lCount < lLength; lCount++)
    {
      lAttrElem = (Element)lAttrElems.item(lCount);
      lAttr     = lAttrElem.getAttribute("ref");

      /***********************************************************************/
      /* All the locally-referenced attributes have refs starting with a '#' */
      /* character.  We won't mess with externally-referenced ones.          */
      /***********************************************************************/
      if (lAttr.startsWith("#"))
      {
        lAttrNames.add(lAttr.substring(1));                    /* remove '#' */
      }
    }

    /*************************************************************************/
    /* Next, iterate through the attribute-types from the main schema; if    */
    /* the id of that attribute-type is in the hash, add it to the mini      */
    /* schema as well.                                                       */
    /*************************************************************************/
    lAttrElems = xiSchemaDoc.getElementsByTagName("attribute-type");
    
    for (lCount = 0; lCount < lAttrElems.getLength(); lCount++)
    {
      lAttrElem = (Element)lAttrElems.item(lCount);
      lAttr = lAttrElem.getAttribute("id");

      if (lAttrNames.contains(lAttr))
      {
        lNode = xbEntryDoc.importNode(lAttrElem, true);
        lSchemaElem.appendChild(lNode); 
      }
    }

    /*************************************************************************/
    /* Now the mini schema has both the correct objectclasses, and the       */
    /* correct attributes for those objectclasses.                           */
    /*************************************************************************/
    
  } /* createMiniSchema() */

/** This method determines whether two sub-trees differ and, if so, may add
    XML comments to that effect.
    This method takes two Elements. It then compares all their children
    with a given tag name, in a sorted order. It does that sort either on
    an XML attribute of those children or on the NodeValue of each child's
    first child. Which it chooses depends on the value of xiAttributeName.
    The thing that it does choose is known by the generic name of a
    "Value".
    If there is a discrepancy, it may (depending on the setting of the
    relevant member variable) add comments.

  @param xiElem1 The first Element to compare
  @param xiElem2 The Element to compare with it
  @param xiTagName The tag name of the sub-elements which we are comparing
  @param xiAttributeName The attribute name of the distinguishing attribute of
    those sub-elements, or null if you want to use the
    NodeValue() of the first child element.
  @param xoSubElements Reference to a hashtable in which to put a hash of all
    sub-elements which match, or null.

  @return True if the elements differ, false if they do not.
  */
  private boolean diffElements(Element   xiElem1, 
                               Element   xiElem2, 
                               String    xiTagName, 
                               String    xiAttributeName,
                               int       xiIndentLevel,
                               Hashtable xoSubElems)
  {
    /*************************************************************************/
    /* Local variables; numbers 1 and 2 refer to xiElem1 and xiElem2         */
    /*************************************************************************/
    Hashtable lHash1      = new Hashtable(); /* hash of sub-elems of xiElem1 */
    Hashtable lHash2      = new Hashtable(); /* hash of sub-elems of xiElem2 */
    Element   lValueElem1 = null;       /* an individual sub-elem of xiElem1 */
    Element   lValueElem2 = null;       /* an individual sub-elem of xiElem2 */

    TreeSet   lSet1      = new TreeSet();  /* Values of sub-elems of xiElem1 */
    TreeSet   lSet2      = new TreeSet();  /* Values of sub-elems of xiElem2 */    
    String    lValue1    = null;  /* individual value of sub-elem of xiElem1 */
    String    lValue2    = null;  /* individual value of sub-elem of xiElem1 */

    Iterator  lIterator1 = null;                       /* Iterator for lSet1 */
    Iterator  lIterator2 = null;                       /* Iterator for lSet2 */
    boolean   lNewValue1 = true;            /* boolean to advance lIterator1 */
    boolean   lNewValue2 = true;            /* boolean to advance lIterator2 */
    Comment   lComment   = null;

    int       lEquals    = 0;              /* whether two Elements are equal */
    boolean   lRetVal    = false;     /* return value - assume same at start */

    /*************************************************************************/
    /* We need an XML serializer to create some of the comments later        */
    /*************************************************************************/
    StringWriter lStrWriter = new StringWriter();
    OutputFormat lOutForm = new OutputFormat(m_Doc2);
    lOutForm.setIndenting(false);
    lOutForm.setIndent(DSMLDocumentImpl.TAB_SIZE);
    lOutForm.setOmitXMLDeclaration(true);
    lOutForm.setLineWidth(79);
    lOutForm.setLineSeparator("\n");
    XMLSerializer lSerializer = new XMLSerializer(lStrWriter, lOutForm);

    /*************************************************************************/
    /* Here, we get all the Element's "Values", both as an ordered set and   */
    /* as an unordered hash with the actual Elements.                        */
    /*************************************************************************/
    getValues(xiElem1, xiTagName, xiAttributeName, lHash1, lSet1);
    getValues(xiElem2, xiTagName, xiAttributeName, lHash2, lSet2);

    /*************************************************************************/
    /* Next, we go through the Values, possibly adding comments if there has */
    /* been a change.                                                        */
    /*************************************************************************/
    lIterator1 = lSet1.iterator();
    lIterator2 = lSet2.iterator();

    lValue1 = lIterator1.hasNext() ? (String)lIterator1.next() : null;
    lValue2 = lIterator2.hasNext() ? (String)lIterator2.next() : null;

    while (lValue1 != null || lValue2 != null)
    {
      lEquals = 0;

      /***********************************************************************/
      /* If we have Values, get the corresponding Elements, or pre-judge the */
      /* comparison to make sure we're not doing operations on a null        */
      /* Element.                                                            */
      /***********************************************************************/
      if (lValue1 != null)
      {
        lValueElem1 = (Element)lHash1.get(lValue1);
      }
      else
      {
        lEquals = 1;
      }

      if (lValue2 != null)
      {
        lValueElem2 = (Element)lHash2.get(lValue2);
      }
      else
      {
        lEquals = -1;
      }

      /***********************************************************************/
      /* Do the comparision between the two Values, but only if lEquals has  */
      /* not been pre-set.                                                   */
      /***********************************************************************/
      if (lEquals == 0)
      {
        lEquals = lValue1.compareTo(lValue2);
      }

      if (lEquals < 0)
      {
        /*********************************************************************/
        /* There is an Value in Elem1 that is not in Elem2.  So, we have to  */
        /* add a comment to Elem2 with the contents of the "missing" Value.  */
        /* We create the comment by serializing the Value from Elem1 to a    */
        /* StringWriter, and then reading the String from it.                */
        /*********************************************************************/
        
        if (m_addComments)
        {
          try
          {
            lSerializer.serialize(lValueElem1);
          }
          catch (IOException lExcept)
          {
          }
  
          String lXMLStr = lStrWriter.toString();
          lXMLStr = indent(lXMLStr, xiIndentLevel);

          lComment = m_Doc2.createComment(" This <" + xiTagName + "> was " +
                                          "removed:\n" + lXMLStr);
          
          (lStrWriter.getBuffer()).setLength(0);
  
          (lValueElem2.getParentNode()).insertBefore(lComment, 
                                                 lValueElem2.getNextSibling());
        }

        lRetVal    = true;

        /*********************************************************************/
        /* If no more, set lValue1 to null                                   */
        /*********************************************************************/
        if (lIterator1.hasNext())
        {
          lValue1 = (String)lIterator1.next();
        }
        else
        {
          lValue1 = null;
        }
      }
      else if (lEquals > 0)
      {
        /*********************************************************************/
        /* There is an Value in Elem2 that is not in Elem1.  So, we need to  */
        /* add a comment to Elem2 saying "This Value is new".                */
        /*********************************************************************/
        
        if (m_addComments)
        {
          lComment = m_Doc2.createComment(" This <" + xiTagName + 
                                          "> is new: ");
          (lValueElem2.getParentNode()).insertBefore(lComment, lValueElem2);
        }

        lRetVal    = true;

        /*********************************************************************/
        /* If no more, set lValue2 to null                                   */
        /*********************************************************************/
        if (lIterator2.hasNext())
        {
          lValue2 = (String)lIterator2.next();
        }
        else
        {
          lValue2 = null;
        }
      }
      else /* (lEquals == 0) */
      {
        /*********************************************************************/
        /* Same Value in Elem1 and Elem2; record this Element and then skip  */
        /* past both.                                                        */
        /*********************************************************************/
        if (xoSubElems != null)
        {
          xoSubElems.put(lValueElem1, lValueElem2);
        }

        /*********************************************************************/
        /* If no more, set lValue1 to null                                   */
        /*********************************************************************/
        if (lIterator1.hasNext())
        {
          lValue1 = (String)lIterator1.next();
        }
        else
        {
          lValue1 = null;
        }

        /*********************************************************************/
        /* If no more, set lValue2 to null                                   */
        /*********************************************************************/
        if (lIterator2.hasNext())
        {
          lValue2 = (String)lIterator2.next();
        }
        else
        {
          lValue2 = null;
        }
      }        
    }

    return lRetVal;

  } /* diffElements() */

/** This method goes through all the descendents of a given element with a
    given tag name, and reads a given XML attribute (or, if lAttrName is
    null, the NodeValue() of the first child.) It then adds the
    attribute/value to a set and also uses it as a key (with the child it
    came from as the data) in a Hashtable.
    Why?  This method allows the calling function to go through descendent
    nodes in sorted order, doing operations on each, if it passes in a set
    which is guaranteed to produce an Iterator which goes through in sorted
    order. It would be simpler if Java provided a class which was a
    Hashtable which could return a sorted-order iterator, but it doesn't.

  @param xiElem The XML element which is the root of this operation
  @param xiTagName The tag name of the children to sort
  @param xiAttrName The name of the XML attribute of the child to sort on, or
    null if you want to use the NodeValue() of the firstChild().
  @param xoHash Used to return the hash of Values and Nodes.
  @param xoSet Used to return the Set of Values.
  */
  private void getValues(Element   xiElem, 
                         String    xiTagName, 
                         String    xiAttrName, 
                         Hashtable xoHash,
                         Set       xoSet) 
  {
    
    /*************************************************************************/
    /* Local variables                                                       */
    /*************************************************************************/
    int     lLength = 0;
    int     lCount  = 0;
    Element lElem   = null;
    Node    lNode   = null;
    String  lValue  = null;

    /*************************************************************************/
    /* We create a hashtable associating all the Values with their Document  */
    /* Elements, and a Sorted Set (TreeSet) with all the Values in.          */
    /*************************************************************************/
    NodeList lElems = xiElem.getElementsByTagName(xiTagName);

    lLength = lElems.getLength();

    /*************************************************************************/
    /* Iterate over all the children                                         */
    /*************************************************************************/
    for (lCount = 0; lCount < lLength; lCount++)
    {
      lElem = (Element)lElems.item(lCount);

      if (xiAttrName != null)
      {
        lValue = lElem.getAttribute(xiAttrName);
      }
      else
      {
        lNode = lElem.getFirstChild();

        /*********************************************************************/
        /* Skip past any comments there might be                             */
        /*********************************************************************/
        while (lNode.getNodeType() == Node.COMMENT_NODE)
        {
          lNode = lNode.getNextSibling();
        }

        lValue = lNode.getNodeValue();

      }

      xoHash.put(lValue, lElem);
      xoSet.add(lValue);
    }

  } /* getValues() */

/** This method "indents" a string by adding a certain number of spaces at
    the beginning, and after any line separator except the last char, if it
    is one.

  @param xiStr The string to indent
  @param xiLevel The level to indent it to; this is multiplied by
    DSMLDocumentImpl.TAB_SIZE
  */
  private String indent(String xiStr, int xiLevel)
  {
    int  lCount  = 0;
    int  lStrLen = xiStr.length();
    char lChar   = '\0';                           /* a character from xiStr */

    StringBuffer lBuf       = new StringBuffer(lStrLen);   /* The new string */
    StringBuffer lSpacesBuf = new StringBuffer(10); /* spaces to indent with */
    String       lSpaces    = null;                 /* spaces to indent with */

    /*************************************************************************/
    /* Create the indent string of the right number of spaces                */
    /*************************************************************************/
    for (lCount = 0; lCount < xiLevel * DSMLDocumentImpl.TAB_SIZE; lCount++)
    {
      lSpacesBuf.append(' ');
    }

    lSpaces = lSpacesBuf.toString();

    /*************************************************************************/
    /* Add the Spaces at the beginning, and before any occurrence of \n,     */
    /* except if the last char is \n.                                        */
    /*************************************************************************/
    lBuf.append(lSpaces);
    
    for (lCount = 0; lCount < (lStrLen - 1); lCount++)
    {
      lChar = xiStr.charAt(lCount);
      lBuf.append(lChar);
      
      if (lChar == '\n')
      {
        lBuf.append(lSpaces);
      }
    }
    
    /*************************************************************************/
    /* Add the last char                                                     */
    /*************************************************************************/
    lChar = xiStr.charAt(lCount);
    lBuf.append(lChar);

    /*************************************************************************/
    /* return the modified string                                            */
    /*************************************************************************/
    return lBuf.toString();

  } /* indent() */

/** This method sets whether human-readable comments are added to the Add
    file.

  @param xiAddComments True to add comments, false to not add comments.
  */
  public void setAddComments(boolean xiAddComments)
  {
    m_addComments = xiAddComments;

  } /* setAddComments() */

} /* DSMLDOMDiff */
