/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.io.*;
import gnu.getopt.*;
import netscape.ldap.*;
import org.w3c.dom.Document;
import org.apache.xerces.dom.DocumentImpl;

/** 

  This class exposes functionality to produce an asymmetric diff of two
  DSML documents.
  <BR><BR>
  This program reads in two DSML files, X and Y and produces two more DSML
  files, D1 and D2, such that:<BR>
  <CODE>X - D1 + D2 = Y</CODE><BR>
  The diff is done at the "directory entry" level, i.e. Take X, remove all
  the entries in D2, add/replace all the entries in D1 and you have Y. Note
  that this is not symmetric, i.e. it is <B>not</B> the case that:<BR>
  <CODE>Y - D2 + D1 = X</CODE>&nbsp;&nbsp;&nbsp;&nbsp;<B>Wrong!</B><BR><BR>
  D1 will have no schema. D2 will have a schema built up from the two
  schemas in X and Y.<BR>
  <BR><PRE>
  The command syntax is as follows. Bold indicates a mandatory option.<BR>
  DSMLDiff <B>infileA -s subfile -a addfile infileB</B> [options]<BR>
  Output options:<UL>
  <LI>--subtract    -s Output file to be subtracted from X.
  <LI>--add         -a Output file to be added to (X - subtractfile).
  <LI>--nocomments  -c Don't add human-readable deletion comments to
  addition file.
  <LI>--quick       -q Quick mode - do not validate DSML documents
  <LI>--help        -? Prints a usage message.
  </UL>
  E.g.:
  DSMLDiff X.dsml Y.dsml -s subtractfile -a addfile</PRE>
  */
public class DSMLDiff extends DSMLFile
{
/** 

  The filename of the from file
  */
  private static String  s_fromfileName = null;

/** 

  The filename of the to file
  */
  private static String  s_tofileName   = null;

/** 

  The filename of the sub file (subtract from the from file)
  */
  private static String  s_subfileName  = null;

/** 

  The filename of the add file (add to (from file - sub file)).
  */
  private static String  s_addfileName  = null;

/** 

  Whether to validate (false) or not (true).
  */
  private static boolean s_quick        = false;

/** 

  Whether to add human-readable comments to the addfile.
  */
  private static boolean s_addComments  = true;

/** This method drives the DSMLDiff process.

  @param xiArgs The command-line
  */
  public static void main(String[] xiArgs)
  {
    /*************************************************************************/
    /* Member variables                                                      */
    /*************************************************************************/
    Document lFromDoc  = null;                        /* The "from" Document */
    Document lToDoc    = null;                          /* The "to" Document */
    String   lComplete = null;   /* Whether a doc has the complete attribute */

    /*************************************************************************/
    /* Parse command line                                                    */
    /*************************************************************************/
    parseCommandLine(xiArgs);

    /*************************************************************************/
    /* Read in the input files                                               */
    /*************************************************************************/
    lFromDoc = readDSML(s_fromfileName);
    lToDoc   = readDSML(s_tofileName);

    /*************************************************************************/
    /* Check for "complete" attribute; if not complete, don't validate.      */
    /*************************************************************************/
    lComplete = (lFromDoc.getDocumentElement()).getAttribute("complete");

    /*************************************************************************/
    /* Validate if required                                                  */
    /*************************************************************************/
    if (lComplete.equals("false"))
    {
      System.err.println("Unable to validate document \"" + s_fromfileName + 
                        "\" as it contains references to an external schema.");
    }
    else if (!s_quick)
    {
      
      /***********************************************************************/
      /* Validate From doc                                                   */
      /***********************************************************************/
      try
      {
        System.err.println("Validating document \"" + s_fromfileName + "\".");
        DSMLDocumentImpl.validate(lFromDoc);
      }
      catch (DSMLRefException lExcept)
      {
        System.err.println("Unable to validate document \"" + s_fromfileName + 
                        "\" as it contains references to an external schema.");
      }
    }

    /*************************************************************************/
    /* Check for "complete" attribute; if not complete, don't validate.      */
    /*************************************************************************/
    lComplete = (lToDoc.getDocumentElement()).getAttribute("complete");

    if (lComplete.equals("false"))
    {
      System.err.println("Unable to validate document \"" + s_tofileName + 
                        "\" as it contains references to an external schema.");
    }
    else if (!s_quick)
    {
      /***********************************************************************/
      /* Validate To doc                                                     */
      /***********************************************************************/
      try
      {
        System.err.println("Validating document \"" + s_tofileName + "\".");
        DSMLDocumentImpl.validate(lToDoc);
      }
      catch (DSMLRefException lExcept)
      {
        System.err.println("Unable to validate document \"" + s_tofileName + 
                        "\" as it contains references to an external schema.");
      }
    }

    /*************************************************************************/
    /* Create two new empty Documents for the sub file and add file          */
    /*************************************************************************/
    DocumentImpl lSubDoc = new DSMLDocumentImpl();
    DocumentImpl lAddDoc = new DSMLDocumentImpl();

    /*************************************************************************/
    /* Do the diff                                                           */
    /*************************************************************************/
    DSMLDOMDiffer lDiffer = new DSMLDOMDiffer();
    lDiffer.setAddComments(s_addComments);

    lDiffer.diff(lFromDoc, 
                 lToDoc, 
                 lSubDoc, 
                 lAddDoc);

    /*************************************************************************/
    /* Serialize the subfile and addfile                                     */
    /*************************************************************************/
    writeDSML(s_subfileName, lSubDoc);
    writeDSML(s_addfileName, lAddDoc);

  } /* main() */

/** This function creates a Getopt object, passing in the argument string,
    and makes repeated calls to Getopt() to set all or some of the
    command-line option member variables.

  @param xiArgs The argument string to parse
  */
  private static void parseCommandLine(String[] xiArgs)
  {
    int lOpt = 0;

    LongOpt[] lLOs = new LongOpt[5];

    lLOs[0]  = new LongOpt("subtract",   LongOpt.REQUIRED_ARGUMENT, null, 's');
    lLOs[1]  = new LongOpt("add",        LongOpt.REQUIRED_ARGUMENT, null, 'a');
    lLOs[2]  = new LongOpt("nocomments", LongOpt.NO_ARGUMENT, null, 'c');
    lLOs[3]  = new LongOpt("quick",      LongOpt.NO_ARGUMENT, null, 'q');
    lLOs[4]  = new LongOpt("help",       LongOpt.NO_ARGUMENT, null, '?');

    Getopt lGetopt = new Getopt("DSMLDiff", 
                                xiArgs, 
                                "s:a:cq?",
                                lLOs);

    String lArg = null;

    while ((lOpt = lGetopt.getopt()) != -1)
    {
      switch (lOpt)
      {
        /*********************************************************************/
        /* Subfile                                                           */
        /*********************************************************************/
        case 's':
          s_subfileName = lGetopt.getOptarg();
          break;

        /*********************************************************************/
        /* Addfile                                                           */
        /*********************************************************************/
        case 'a':
          s_addfileName = lGetopt.getOptarg();
          break;


        /*********************************************************************/
        /* Turn off commenting?                                              */
        /*********************************************************************/
        case 'c':
          s_addComments = false;
          break;
        
        /*********************************************************************/
        /* Turn off validation?                                              */
        /*********************************************************************/
        case 'q':
          s_quick = true;
          break;

        /*********************************************************************/
        /* Print usage message?                                              */
        /*                                                                   */
        /* This value can be requested on the command line, but is also      */
        /* returned when Getopt encounters an error.                         */
        /*********************************************************************/
        case '?':
          printUsageMessage();
          System.exit(0);
          break;
        
        /*********************************************************************/
        /* Default - invalid option                                          */
        /*********************************************************************/
        default:
          System.err.println("DSMLDiff: Invalid option: " + lOpt);
          printUsageMessage();
          System.exit(1);
          break;
      }
    }

    /*************************************************************************/
    /* Get non-option command line parameters.  There should be two of       */
    /* these.  Optind is left pointing at them.                              */
    /*************************************************************************/
    int lIndex = lGetopt.getOptind();
    
    try
    {
      s_fromfileName = xiArgs[lIndex++];
      
      s_tofileName = xiArgs[lIndex++];
    }
    catch (ArrayIndexOutOfBoundsException lExcept)
    {
      /***********************************************************************/
      /* If there are no non-option arguments, getopt sets Optind to one     */
      /* past the end of the array (not shockingly helpfully.) So, this      */
      /* exception means that no outfileName was specified; we'll use        */
      /* stdout.                                                             */
      /***********************************************************************/
      System.err.println("DSMLDiff: You must specify from-file and to-file.");
      printUsageMessage();
      System.exit(1);
    }

    /*************************************************************************/
    /* If we haven't fallen off the end of xiArgs yet, there are too many    */
    /* command-line parameters.                                              */
    /*************************************************************************/
    if (lIndex < xiArgs.length)
    {
      System.err.println("DSMLDiff: Too many filenames specified.");
      printUsageMessage();
      System.exit(1);
    }

    /*************************************************************************/
    /* Check for mandatory attributes                                        */
    /*************************************************************************/
    if (s_subfileName == null)
    {
      System.err.println("DSMLDiff: You must supply a sub file (-s).");
      printUsageMessage();
      System.exit(1);
    }    

    if (s_addfileName == null)
    {
      System.err.println("DSMLDiff: You must supply an add file (-a).");
      printUsageMessage();
      System.exit(1);
    }    

  } /* parseCommandLine() */

  /***************************************************************************/
  /* Method: printUsageMessage()                                             */
  /*                                                                         */
  /* This method prints a usage message - it should be called if the user    */
  /* incorrectly specifies one or more command-line parameters.              */
  /***************************************************************************/
  private static void printUsageMessage()
  {
    System.err.print(
    "\n" +
    "This command produces <subfile> and <addfile> such that:\n" +
    "<infileA> - <subfile> + <addfile> = <infileB>\n" +
    "with addition being in overwrite mode.\n" +
    "\n" +
    "The command syntax is as follows (with commands in standard Unix format):\n" +
    "DSMLDiff <infileA> -s <subfile> -a <addfile> <infileB> [options]\n" +
    "\n" +
    "Output options:\n" +
    "--subtract   -s  Output file (to be subtracted from A.)\n" +
    "--add        -a  Output file (to be added to (A - <subfile>).)\n" +
    "--nocomments -c  Don't add human-readable deletion comments to add file.\n" +
    "--quick      -q  Quick mode - do not validate DSML documents\n" +
    "--help       -?  Prints a usage message.\n" +
    "\n"
    );
  }

  /***************************************************************************/
  /* TEST_CODE                                                               */
  /***************************************************************************/
  private static void printMemberVariables()
  {      
    System.err.print("s_fromfileName: ");
    System.err.println(s_fromfileName);
    System.err.print("s_tofileName: ");
    System.err.println(s_tofileName);
    System.err.print("s_addfileName: ");
    System.err.println(s_addfileName);
    System.err.print("s_subfileName: ");
    System.err.println(s_subfileName);
    System.err.print("s_addComments: ");
    System.err.println(s_addComments);
    System.err.print("s_quick: ");
    System.err.println(s_quick);
  }

} /* DSMLDiff */

