/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import org.apache.xerces.dom.DocumentImpl;
import org.apache.xml.serialize.*;
import netscape.ldap.*;

/** 

  This class extends DocumentImpl, an implementation of the Document
  interface, to add DSML-specific functionality, such as directory-style
  validation.
  */
public final class DSMLDocumentImpl extends DocumentImpl
{
/** 

  The amount each line is indented from the previous one when a
  DSMLDocumentImpl is serialized.
  */
  public static final int TAB_SIZE = 2;

/** When we create new DSML documents, this constructor ensures that they
    will automatically have the &lt;dsml&gt; node.

  */
  public DSMLDocumentImpl()
  {
    super();

    /*************************************************************************/
    /* Add the <dsml> node                                                   */
    /*************************************************************************/
    Node lDocNode = this.appendChild(this.createElement("dsml"));
    ((Element)lDocNode).setAttribute("xmlns", "http://www.dsml.org/DSML");

  } /* DSMLDocumentImpl() */

/** This function calls validate(this, null) on this Document.

  */
  public void validate() throws DSMLRefException
  {
    validate(this, null);

  } /* validate () */

/** This function calls validate(xiDoc, null) on this Document - in other
    words, validation using the Document's schema.

  @param xiDoc The Document to validate
  */
  public static void validate(Document xiDoc) throws DSMLRefException
  {
    validate(xiDoc, null);

  } /* validate () */

/** This function checks the directory data in the passed in Document for
    consistency at a directory (rather than XML) level. Specifically, it
    checks that the following are true:
    <UL>
    <LI>The schema only contains attributes referenced in one or more
    objectclasses
    <LI>Each entry has only objectclasses present in the schema
    <LI>Each entry has only the attributes permitted by its objectclasses
    <LI>Each entry has all the mandatory attributes for each of its
    objectclasses
    </UL>
    The structure of a DSML document is best ascertained from the DSML
    spec, but in summary: The schema is denoted by two sorts of XML tag -
    &lt;class&gt;, representing an objectclass, and &lt;attribute-type&gt;,
    representing the definition of an attribute. The data is enclosed in
    &lt;entry&gt; tags.<BR><BR>
    It takes the following actions on finding invalid bits of DSML:
    <UL>
    <LI>&lt;attribute-type&gt; referenced in &lt;class&gt; but not present
    - dies with error (very serious problem).
    <LI>&lt;attribute-type&gt; unreferenced by &lt;class&gt; - removes
    &lt;attribute-type&gt;.
    <LI>&lt;oc-value&gt; references &lt;class&gt; which is not present -
    removes &lt;oc-value&gt;.
    <LI>Required &lt;attr&gt; missing from &lt;entry&gt; - removes
    &lt;entry&gt;.
    <LI>&lt;attr&gt; in entry not permitted by objectclasses - removes
    &lt;attr&gt;.
    </UL>

  In some situations (DSML2LDAP) the schema information we are validating
  against is in LDAP structure form, and in others it's in XML form, as
  part of the same document as the data. This function copes with both,
  taking a LPDASchema object as a parameter. If this is null, it uses the
  schema information in the Document instead.  Doing things this way
  prevents converting a massive schema into Document form, passing it
  into validate() and then converting half of it back to work with.<BR>
  <BR>
  We assume any schema we are given in LDAPSchema form has been retrieved
  from a directory, and so don't validate it. (It's in the wrong form
  to do so, anyway.)

  @param xiDoc The Document to validate
  @param xiSchema An LDAPSchema object representing the schema to validate
    against, or null to use the Document's schema.

  @exception DSMLRefException Thrown if, in the course of validation, we encounter
    a reference to an external schema. This makes
    continuing with validation impossible, as we might
    remove valid data.
  */
  public static void validate(Document xiDoc, LDAPSchema xiSchema)
                              throws DSMLRefException
  {
    LDAPSchema lSchema = null;

    /*************************************************************************/
    /* Assume schema we are given in LDAP form are from a directory, and so  */
    /* don't validate them.  However, we validate schemas from Documents.    */
    /*************************************************************************/
    if (xiSchema == null)
    {
      NodeList lSchemaNodes = xiDoc.getElementsByTagName("directory-schema");
      
      if (lSchemaNodes.getLength() > 0)
      {
        /*********************************************************************/
        /* We are using the schema from the DSML Document, and it exists.    */
        /* Validate it (first bullet).                                       */
        /*********************************************************************/
        validateSchema(xiDoc);  

        /*********************************************************************/
        /* Convert this validated schema to an LDAPSchema so we can validate */
        /* the entries.                                                      */
        /*********************************************************************/
        lSchema = DSMLLDAPConverter.nodeToSchema(lSchemaNodes.item(0));
      }
      else
      {
        System.err.println("Cannot validate - document has no schema.");
      }
    }
    else
    {
      /*********************************************************************/
      /* We are using the schema passed in                                 */
      /*********************************************************************/
      lSchema = xiSchema;
    }                                         

    /***********************************************************************/
    /* Entry has valid objectclasses and attributes (second, third and     */
    /* fourth bullets)                                                     */
    /***********************************************************************/
    if (lSchema != null)
    {
      validateEntries(xiDoc, lSchema);
    }

  } /* validate() */

/** Check that the schema only contains attributes referenced in one or
    more local objectclasses, and that all local attributes referenced in
    objectclasses are present.

  @param xiDoc The Document object whose schema we are validating
  */
  private static void validateSchema(Document xiDoc) throws DSMLRefException
  {
    /*************************************************************************/
    /* Local variables                                                       */
    /*************************************************************************/
    int      lCount    = 0;     
    NodeList lList     = null;
    String   lName     = null;
    Iterator lIterator = null;

    Element  lAttrElem   = null;                        /* an <attr> element */
    HashSet  lAttrNameSet  = new HashSet();    /* a list of legal attributes */ 
    HashSet  lFoundNameSet = new HashSet();              /* Valid attributes */

    /*************************************************************************/
    /* Iterate over all the &lt;attribute&gt; tags in the document.          */
    /* Attribute tags are children of &lt;class&gt; tags, and represent the  */
    /* legal attributes for each objectclass.                                */
    /*************************************************************************/
    lList = xiDoc.getElementsByTagName("attribute");

    for (lCount = 0; lCount < lList.getLength(); lCount++)
    {
      /***********************************************************************/
      /* Work out the attribute URI ref and add it to the set.  Note that    */
      /* this is not necessarily the same thing as the attribute's NAME, as  */
      /* long as it matches the id attribute of the &lt;attribute-type&gt;.  */
      /***********************************************************************/
      lName = ((Element)lList.item(lCount)).getAttribute("ref");
      if (lName.startsWith("#"))
      {
        lName = lName.substring(1);                            /* remove '#' */
        lAttrNameSet.add(lName);
      }
      else
      {
        throw new DSMLRefException("Non-local URI ref in DSML file.");
      }
    }

    /*************************************************************************/
    /* Now, iterate over the attributes and see if they are referenced by an */
    /* objectclass                                                           */
    /*************************************************************************/
    lList = xiDoc.getElementsByTagName("attribute-type");

    for (lCount = 0; lCount < lList.getLength(); lCount++)
    {
      /***********************************************************************/
      /* For each attribute, we need to find the id.                         */
      /***********************************************************************/
      lAttrElem = (Element)lList.item(lCount);
      lName = lAttrElem.getAttribute("id");
      
      if (lAttrNameSet.contains(lName))
      {
        /*********************************************************************/
        /* We've found the name. We can't remove it yet, as attributes may   */
        /* be referenced by more than one objectclass.                       */
        /*********************************************************************/
        lFoundNameSet.add(lName);
      }
      else
      {
        /*********************************************************************/
        /* This attribute is not referenced by any objectclass - so remove   */
        /* it from the tree.                                                 */
        /*********************************************************************/
        System.err.println("Attribute \"" + lName + "\" is not referenced " +
                           "by any objectclass - removing from schema.");
        (lAttrElem.getParentNode()).removeChild(lAttrElem);

        /*********************************************************************/
        /* Removing this changes the index in the NodeList, so we have to do */
        /* lCount-- to keep in sync.                                         */
        /*********************************************************************/
        lCount--;
      }
    }

    /*************************************************************************/
    /* Now, remove all the FoundNames from the lAttrNameSet                  */
    /*************************************************************************/
    lIterator = lFoundNameSet.iterator();
    
    while (lIterator.hasNext())
    {
      lName = (String)lIterator.next();
      lAttrNameSet.remove(lName);
    }

    /*************************************************************************/
    /* Have we found all the local attributes referenced by the              */
    /* objectclasses?  If not, this is an error we can't fix.                */
    /*************************************************************************/
    if (!lAttrNameSet.isEmpty())
    {
      lIterator = lAttrNameSet.iterator();

      while (lIterator.hasNext())
      {
        lName = (String)lIterator.next();

        System.err.println("Attribute \"" + lName + "\" is referenced as " +
                           "local by an objectclass but is not present in " +
                           "the schema. I cannot fix this problem.");
        System.exit(1);
      }
    }

  } /* validateSchema() */

/** Check that each entry has valid objectclasses, only the attributes
    permitted by those objectclasses, and contains all mandatory
    attributes.
    Because LDAP is case-insensitive, we use lower case for the lAttrTable
    internally, while keeping the strings upper case as we go through, for
    cosmetic reasons. This may mean we convert the same string to lower
    case multiple times.

  @param xiDoc The document we are validating
  @param xiOCTable A hashtable of LDAPObjectClassSchema objects and their names
  */
  private static void validateEntries(Document   xiDoc,
                                      LDAPSchema xiSchema) 
  {
    /*************************************************************************/
    /* Local variables; numerical suffixes refer to nested loops over        */
    /* NodeLists.                                                            */
    /*************************************************************************/
    NodeList    lList      = null;
    NodeList    lList2     = null;
    NodeList    lList3     = null;
    int         lCount     = 0;     
    int         lCount2    = 0;
    int         lCount3    = 0;
    int         lLength    = 0;
    int         lLength2   = 0;
    int         lLength3   = 0;

    String      lName      = null;
    String      lDN        = null;
    Node        lNode      = null;
    Node        lTextNode  = null;
    Iterator    lIterator  = null;
    Enumeration lEnum      = null;

    LDAPObjectClassSchema lOCSchema = null;

    Element   lEntryElem   = null;                     /* an <entry> Element */
    String    lEntryName   = null;               /* the name of that Element */
    boolean   lKilledEntry = false;     /* marks that that an entry must die */ 

    Element   lAttrElem = null;                         /* an <attr> Element */

    Hashtable lAttrTable = new Hashtable(); /* Attribute nodes of an entry,  */
                                            /* with the names as keys        */
    HashSet  lOCNameSet    = new HashSet(); /* names of OCs used by an entry */
    HashSet  lFoundNameSet = new HashSet(); /* Attribute names we find in an */
                                            /* entry                         */
    

    /*************************************************************************/
    /* Grab all the entries in the Document                                  */
    /*************************************************************************/
    lList = xiDoc.getElementsByTagName("entry");

    /*************************************************************************/
    /* For each entry:                                                       */
    /*************************************************************************/
    for (lCount = 0; lCount < lList.getLength(); lCount++)
    {
      lEntryElem = (Element)lList.item(lCount);
      lEntryName = lEntryElem.getAttribute("dn");
      lKilledEntry = false;

      /***********************************************************************/
      /* We need to make a list of the names of its attributes and its       */
      /* objectclasses.  This is so we can go through the objectclasses and  */
      /* check the attributes are the right ones.  We do this by going       */
      /* through the &lt;entry&gt;'s children looking for Element nodes.     */
      /***********************************************************************/
      lList2   = lEntryElem.getChildNodes();
      lLength2 = lList2.getLength();

      for (lCount2 = 0; lCount2 < lLength2; lCount2++)
      {
        lNode = lList2.item(lCount2);

        /*********************************************************************/
        /* We are only working with Element nodes                            */
        /*********************************************************************/
        if (lNode.getNodeType() == Node.ELEMENT_NODE)
        {          
          lAttrElem = (Element)lNode;
  
          /*******************************************************************/
          /* Try to get the "name" XML attribute                             */
          /*******************************************************************/
          lName = lAttrElem.getAttribute("name");
  
          if (!lName.equals(""))
          {
            /*****************************************************************/
            /* Only attributes have "name" XML attributes.  We are storing   */
            /* the names of the attributes, together with the Nodes          */
            /* representing them, in a Hashtable.                            */
            /*****************************************************************/
            lAttrTable.put(lName.toLowerCase(), lAttrElem);
          }
          else
          {
            /*****************************************************************/
            /* Found an objectclass (no "name" attribute)                    */
            /*****************************************************************/
            String lNodeName = lAttrElem.getNodeName();
            
            boolean lIsOC = lNodeName.equalsIgnoreCase("objectclass");
                        
            /*****************************************************************/
            /* We add this to lAttrTable as well, because otherwise we think */
            /* that "required attribute 'objectclass' isn't present."        */
            /*****************************************************************/
            lAttrTable.put(lNodeName, lAttrElem);
            
            /*****************************************************************/
            /* Extract the objectclasses to which this entry belongs, and    */
            /* add them to the list.                                         */
            /*****************************************************************/
            lList3   = lAttrElem.getElementsByTagName("oc-value");
            lLength3 = lList3.getLength();
            
            for (lCount3 = 0; lCount3 < lLength3; lCount3++)
            {
              Node lOCNode = lList3.item(lCount3);

              String lOCName = DSMLLDAPConverter.getSubText(lOCNode);

              /***************************************************************/
              /* Make sure this name is actually a valid objectclass in the  */
              /* schema                                                      */
              /***************************************************************/
              lOCSchema = xiSchema.getObjectClass(lOCName);

              if (lOCSchema == null)
              {
                /*************************************************************/
                /* objectclass not present in schema - remove from entry     */
                /*************************************************************/
                Node lEntryNode = lAttrElem.getParentNode(); 
                lDN = ((Element)lEntryNode).getAttribute("dn");
                System.err.println("Objectclass " + lOCName + " of entry " +
                                   "with dn \"" + lDN + "\" not in " + 
                                   "schema - removing <oc-value>.");

                lAttrElem.removeChild(lOCNode);        
              }
              else
              {
                /*************************************************************/
                /* Yes, it's valid - add it to the list of objectclasses     */
                /* this entry possesses                                      */
                /*************************************************************/
                lOCNameSet.add(lOCName);
              }
            }
          }
        }
      }

      /***********************************************************************/
      /* Now, we check the entry to see if it has all the necessary          */
      /* attributes, and no illegal ones.                                    */
      /*                                                                     */
      /* For each objectclass:                                               */
      /***********************************************************************/
      lIterator = lOCNameSet.iterator();

      while ((lKilledEntry == false) && lIterator.hasNext())
      {
        /*********************************************************************/
        /* For each one, get the LDAPObjectClassSchema object                */
        /*********************************************************************/
        lName = (String)lIterator.next();
        lOCSchema = xiSchema.getObjectClass(lName);

        /*********************************************************************/
        /* Check to see if all required attributes are present               */
        /*********************************************************************/
        lEnum = lOCSchema.getRequiredAttributes();

        while ((lKilledEntry == false) && lEnum.hasMoreElements())
        {
          lName = (String)lEnum.nextElement();
  
          if (lAttrTable.containsKey(lName.toLowerCase()))
          {
            /*****************************************************************/
            /* As we find each attribute, add it to the found name set       */
            /*****************************************************************/
            lFoundNameSet.add(lName);
          }
          else
          {
            /*****************************************************************/
            /* This entry does not have a required attribute.  The data is   */
            /* therefore invalid and the entry must be discarded.            */
            /*****************************************************************/
            System.err.println("Required attribute \"" + lName + "\" is " +
                               "missing from entry with dn \"" + lEntryName +
                               "\". Removing entry."); 
            (lEntryElem.getParentNode()).removeChild(lEntryElem); 
            lCount--;
            lKilledEntry = true;
          }
        }

        /*********************************************************************/
        /* Check to see which optional attributes are present                */
        /*********************************************************************/
        lEnum = lOCSchema.getOptionalAttributes();

        while ((lKilledEntry == false) && lEnum.hasMoreElements())
        {
          lName = (String)lEnum.nextElement();

          if (lAttrTable.containsKey(lName.toLowerCase()))
          {
            /*****************************************************************/
            /* As we find each attribute, add it to the FoundNameSet         */
            /*****************************************************************/
            lFoundNameSet.add(lName);
          }
        }
      }

      if (!lKilledEntry)
      {
        /*********************************************************************/
        /* Remove the found names from the Table; we can't do this earlier   */
        /* because multiple objectclasses may have the same mandatory        */
        /* attributes.                                                       */
        /*********************************************************************/
        Iterator lIterator2 = lFoundNameSet.iterator();
  
        while (lIterator2.hasNext())
        {
          lName = (String)lIterator2.next(); 
          lAttrTable.remove(lName.toLowerCase());
        }
  
        if (!lAttrTable.isEmpty())
        {
          /*******************************************************************/
          /* There are attributes which are not permitted for any of the     */
          /* objectclasses of which this entry is a member.  Remove them.    */
          /*******************************************************************/
          lEnum = lAttrTable.keys();
  
          while (lEnum.hasMoreElements())
          {
            lName = (String)lEnum.nextElement();
            lAttrElem = (Element)(lAttrTable.remove(lName.toLowerCase()));
            System.err.println("Attribute \"" + lName + "\" is not permitted" +
                               " by the objectclasses of entry with dn \"" 
                               + lEntryName + "\". Removing attribute.");
            
            lEntryElem.removeChild(lAttrElem);
          }
        }
      }

      /***********************************************************************/
      /* Clean up for next entry                                             */
      /***********************************************************************/
      lAttrTable.clear();
      lOCNameSet.clear();
      lFoundNameSet.clear();
    }

  } /* validateEntries() */
                                                                            
/** This method serializes this Document to the given OutputStream.

  @param xiDoc Document to serialize
  */
  public void serialize(OutputStream xiOut)
  {
    serialize(this, xiOut);

  } /* serialize() */

/** This method takes a Document and writes its contents as a stream of
    DSML to the given OutputStream.

  @param xiDoc Document to serialize
  @param xiOut OutputStream to serialize the document to
  */
  public static void serialize(Document xiDoc, OutputStream xiOut)
  {
    /*************************************************************************/
    /* Create a serializer (org.apache.xml.serialize.XMLSerializer)          */
    /*************************************************************************/
    try 
    {
      OutputStreamWriter lWriter = new OutputStreamWriter(xiOut, "UTF8");

      OutputFormat lOutForm = new OutputFormat(xiDoc);
      lOutForm.setIndenting(true);
      lOutForm.setIndent(TAB_SIZE);
      lOutForm.setLineWidth(79);
      lOutForm.setLineSeparator("\n");

      XMLSerializer lSerializer = new XMLSerializer(lWriter, lOutForm);
      /***********************************************************************/
      /* Give it the document                                                */
      /***********************************************************************/
      lSerializer.serialize(xiDoc);
    }
    catch (Exception lExcept)
    {
      System.err.println("Exception while serializing: " + 
                         lExcept.toString());
    }

  } /* serialize() */

} /* DSMLDocumentImpl */
                                                                            
