/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.io.*;
import org.apache.xerces.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import org.w3c.dom.traversal.*;
import org.apache.xerces.dom.*;
import org.dsmltools.util.WhitespaceFilter;

/** 

  This class does file IO for DSML documents.  It contains methods to read
  a file containing DSML from disk, and write a Document out as a DSML
  file.
  */
public class DSMLFile
{
/** Reads DSML from a given filename, or stdin if the filename is null.

  @param xiFilename The filename to read from, or null to read from stdin.

  @return The Document which was read
  */
  protected static Document readDSML(String xiFilename)
  {
    DOMParser   lParser      = null;
    Document    lDoc         = null;
    InputSource lInputSource = null;

    /*************************************************************************/
    /* Create the InputSource                                                */
    /*************************************************************************/
    try
    {
      if (xiFilename != null)
      {
        // Make sure the stream is read in as UTF-8
        FileInputStream   lStream = new FileInputStream(xiFilename);
        InputStreamReader lReader = new InputStreamReader(lStream, "UTF8");
        lInputSource              = new InputSource(lReader);
      }
      else
      {
        lInputSource = new InputSource(System.in);
      }
    }
    catch (FileNotFoundException lExcept)
    {
      System.err.println("Input file \"" + xiFilename + "\" not found.");
      System.exit(1);
    }
    catch (UnsupportedEncodingException lExcept)
    {
      System.err.println("Unsupported encoding: " + lExcept.getMessage());
      System.exit(1);
    }

    if (lInputSource == null)
    {
      System.err.println("Unable to create an InputSource.");
      System.exit(1);
    }

    /*************************************************************************/
    /* Create a parser                                                       */
    /*************************************************************************/
    lParser = new DOMParser();

    try
    {
    //  lParser.setIncludeIgnorableWhitespace(false);
      lParser.setFeature(
             "http://apache.org/xml/features/dom/defer-node-expansion", false);

    }
    catch (Exception lExcept)
    {
    }

    /*************************************************************************/
    /* Read in the DSML                                                      */
    /*************************************************************************/
    try
    {
      lParser.parse(lInputSource);      
    }
    catch (IOException lExcept)
    {
      System.err.println("IOException while parsing: " + lExcept.toString());
      System.exit(1);
    }
    catch (SAXException lExcept)
    {
      System.err.println("SAXException while parsing: " + 
                         lExcept.toString());
      System.exit(1);
    }

    /*************************************************************************/
    /* Get the Document from the parser                                      */
    /*************************************************************************/
    lDoc = lParser.getDocument();

    if (lDoc == null)
    {
      System.err.println("Parser didn't return a document.");
      System.exit(1);
    }

    /*************************************************************************/
    /* Remove ignorable whitespace.  We do this by iterating over all text   */
    /* nodes, and removing ones which pass the WhitespaceFilter.             */
    /*                                                                       */
    /* [ISSUE] It should be possible to get the parser to not add ignorable  */
    /* whitespace into the tree, but I haven't worked out how to do that     */
    /* yet.                                                                  */
    /*************************************************************************/
    Node lDocNode = (Node)lDoc.getDocumentElement();
    WhitespaceFilter lWhitespaceFilter = new WhitespaceFilter();

    NodeIterator lNIterator;
    lNIterator = ((DocumentImpl)lDoc).createNodeIterator(lDocNode,
                                                         NodeFilter.SHOW_TEXT,
                                                         lWhitespaceFilter,
                                                         false);

    /*************************************************************************/
    /* All nodes returned by this iterator are whitespace text nodes         */
    /*************************************************************************/
    Node lNode = lNIterator.nextNode();

    while (lNode != null)
    {
      /***********************************************************************/
      /* Remove them                                                         */
      /***********************************************************************/
      (lNode.getParentNode()).removeChild(lNode);
      lNode = lNIterator.nextNode();
    }

    return lDoc;

  } /* readDSML() */

/** Writes a given document to a given filename, or to stdout if the
    filename is null.

  @param xiFileName The file to write to, or null for stdout
  @param xiDoc The Document to write
  */
  protected static void writeDSML(String xiFileName, Document xiDoc)
  {
    if (xiFileName != null)
    {
      try
      {
        FileOutputStream lOutStream = new FileOutputStream(xiFileName);

        DSMLDocumentImpl.serialize(xiDoc, lOutStream);
      }
      catch (FileNotFoundException lExcept)
      {      
        System.err.println("Unable to create output file \"" + xiFileName + 
                           "\".");
        System.exit(1);
      }
    }
    else
    {
      /***********************************************************************/
      /* stdout is the default                                               */
      /***********************************************************************/
      DSMLDocumentImpl.serialize(xiDoc, System.out);
    }

  } /* write() */

} /* DSMLFile */
