/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.util.*;
import netscape.ldap.*;
import org.w3c.dom.*;

/** 

  This class implements functionality to search an LDAP directory and
  return the results of the search as a DSML document.
  */
public final class DSMLLDAPReader extends DSMLLDAPConnection 
{
/** 

  Constants for use in setting the value of m_contents
  */
  public static final int SCHEMA = 1;
  public static final int DATA   = 2;
  public static final int ALL    = 3;                       /* SCHEMA | DATA */

/** 

  Flag to indicate, when true, that we output the entire directory schema
  as DSML, rather than just the parts referenced by the data we are
  outputting.
  */
  private boolean m_outputFullSchema = false;
  
/** 

  Tri-state flag to indicate whether we want the schema, the data or both
  in the output DSML.
  */
  private int     m_contents = ALL;

/** This method has an interface like that of LDAPConnection.search, except
    that it returns a Document object, representing the directory
    information as DSML, rather than an LDAPSearchResults object.

  @param xiBaseDN The base DN for the search
  @param xiScope The scope of the search
  @param xiFilter The filter to use
  @param xiAttributes The attributes to return (default all)
  @param xiAttrsOnly Whether or not to return only attributes and not ocs.
  @param xiCons A structure defining any constraints on the search

  @return A DSMLDocumentImpl containing the data returned by the search, and
  @return information as requested
  */
  public DSMLDocumentImpl search(String                 xiBaseDN, 
                                 int                    xiScope, 
                                 String                 xiFilter,
                                 String[]               xiAttributes, 
                                 boolean                xiAttrsOnly, 
                                 LDAPSearchConstraints  xiCons)
  {
    /*************************************************************************/
    /* Local variables                                                       */
    /*************************************************************************/
    Iterator    lIterator = null;                  
    Enumeration lEnum     = null;
    Node        lNode     = null;
    int         lCount    = 0;

    Element   lEntriesElem  = null;       /* the <directory-entries> Element */
    Vector    lEntries      = new Vector();       /* a vector of LDAPEntries */
    LDAPEntry lEntry        = null;                    /* a single LDAPEntry */

    if (m_conn == null) 
    {
      System.err.println("Internal error: search() when not connected.");
      System.exit(1);
    }

    LDAPSearchResults lResults = null;

    /*************************************************************************/
    /* If we are doing a mini schema, we need to make sure that we get the   */
    /* objectclasses of the search results.  If the user has supplied a list */
    /* of attributes, and it doesn't include "objectclass", we need to add   */
    /* it in.  This assumes the directory won't care if we specify the same  */
    /* attribute twice.                                                      */
    /*************************************************************************/
    if (!(m_outputFullSchema || (m_contents & SCHEMA) == 0))
    {
      if (xiAttributes != null)
      {
        
        String[] lNewAttrs = new String[xiAttributes.length + 1];
        lNewAttrs[0] = new String("objectclass");
        
        for (lCount = 0; lCount < xiAttributes.length; lCount++)
        {
          lNewAttrs[lCount + 1] = xiAttributes[lCount];
        }

        xiAttributes = lNewAttrs;
      }
    }

    /*************************************************************************/
    /* Do the search                                                         */
    /*************************************************************************/
    if ((m_contents & DATA) != 0)
    {
      try
      {
        lResults = m_conn.search(xiBaseDN, 
                                 xiScope, 
                                 xiFilter, 
                                 xiAttributes, 
                                 xiAttrsOnly,
                                 xiCons);
      } 
      catch (LDAPException lExcept) 
      {
        switch (lExcept.getLDAPResultCode())
        {
          case LDAPException.NO_SUCH_OBJECT:
            System.err.println("No entries match your search parameters.");
            System.exit(0);
            break;
        
          default:
            System.err.println("LDAPException during search: " + 
                               lExcept.toString());
            System.exit(1);
            break;
        }
      }
    }

    /*************************************************************************/
    /* For some reason, the search above can, in some cirumstances, stuff up */
    /* the connection. Therefore, looking for the schema will just hang.     */
    /* This reconnect() seems to solve the problem.                          */
    /*************************************************************************/
    try
    {
      m_conn.reconnect();
    }
    catch (LDAPException lExcept)
    {
      System.err.println("LDAPException during reconnect: " +
                         lExcept.toString());
    }

    /*************************************************************************/
    /* Get the full schema from the directory; we have to do this even if we */
    /* aren't outputting it because we need it to check if &lt;attr&gt;      */
    /* values are text or binary.                                            */
    /*************************************************************************/
    LDAPSchema lFullSchema = new LDAPSchema();

    try
    {
      lFullSchema.fetchSchema(m_conn);
    }
    catch (LDAPException lExcept)
    {
      switch(lExcept.getLDAPResultCode()) 
      {
        case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
          System.err.println("Insufficient access rights to read schema.");
          m_contents = m_contents & ~SCHEMA;
          break;
        
        default:
          System.err.println("LDAPException while fetching schema: " 
                             + lExcept.toString());
          break;
      }
    }

    /*************************************************************************/
    /* Create a skeleton DSML Document                                       */
    /*************************************************************************/
    DSMLDocumentImpl lDoc = new DSMLDocumentImpl();

    /*************************************************************************/
    /* Iterate over the returned entries and store as a Vector.  We have to  */
    /* do this if we are creating a mini schema, even if we are not          */
    /* outputting the data.                                                  */
    /*************************************************************************/
    if ((m_contents & DATA) != 0 || !m_outputFullSchema)
    {
      
      /***********************************************************************/
      /* Make sure we've got some results                                    */
      /***********************************************************************/
      if ((lResults == null) || (lResults.hasMoreElements() == false))
      {
        System.err.println("Search returned no results.");
        System.exit(1);
      }
      else
      {
        while (lResults.hasMoreElements()) 
        {                          
          try 
          {                                                    
            lEntry = lResults.next();                      
          } 
          catch (LDAPReferralException lExcept)
          { 
            System.err.println("LDAPReferralException getting next result: ");
            System.err.println(lExcept.toString());
            continue;                                                     
          } 
          catch (LDAPException lExcept) 
          {                            
            switch(lExcept.getLDAPResultCode()) 
            {
              case LDAPException.SIZE_LIMIT_EXCEEDED:
                /*************************************************************/
                /* This isn't a problem                                      */
                /*************************************************************/
                System.err.println("Not showing all results" + 
                     " (use -z to set max results shown.)");
                break;
              
              default:
                System.err.println("LDAPException while looking at results: ");
                System.err.println(lExcept.toString());
                break;
            }
  
            lEntry = null;
          }                                                        
                                                                        
          /*******************************************************************/
          /* Add to Vector                                                   */
          /*******************************************************************/
          if (lEntry != null)
          {
            lEntries.add(lEntry);
          }
        }
      } 
    }

    /*************************************************************************/
    /* If we are outputting the data, we convert the entries from LDAP to    */
    /* Document representation, adding them to the Document as we go.        */
    /*************************************************************************/
    if ((m_contents & DATA) != 0)
    {
      lEntriesElem = lDoc.createElement("directory-entries");
      (lDoc.getDocumentElement()).appendChild(lEntriesElem);

      lIterator = lEntries.iterator();
  
      while (lIterator.hasNext())
      {
        lNode = DSMLLDAPConverter.entryToNode((LDAPEntry)lIterator.next(), 
                                              lDoc, 
                                              lFullSchema);
  
        lEntriesElem.appendChild(lNode);
      }
    }

    /*************************************************************************/
    /* We only process the schema if it is needed for the output.            */
    /*************************************************************************/
    if ((m_contents & SCHEMA) != 0)
    {
      /***********************************************************************/
      /* Which schema is required?                                           */
      /***********************************************************************/
      LDAPSchema lOutSchema = lFullSchema;
  
      if (!m_outputFullSchema)
      {
        /*********************************************************************/
        /* Create a mini schema; lEntries will have been filled              */
        /*********************************************************************/
        lOutSchema = createMiniSchema(lEntries, lFullSchema);
      }
  
      if (lOutSchema != null)
      {
        /*********************************************************************/
        /* We convert the schema from LDAP to Document representation,       */
        /* adding it to the Document as we go.  First, we create the         */
        /* directory-schema root element.                                    */
        /*********************************************************************/
        Element lSchemaElem = null;
  
        lSchemaElem = lDoc.createElement("directory-schema");

        /*********************************************************************/
        /* (If we aren't doing entries, lEntriesElem will be null, and so    */
        /* lSchemaElem will just be inserted at the end of the list of       */
        /* children.)                                                        */
        /*********************************************************************/
        (lDoc.getDocumentElement()).insertBefore(lSchemaElem, lEntriesElem);
    
        /*********************************************************************/
        /* Iterate through the objectclasses, adding them...                 */
        /*********************************************************************/
        LDAPObjectClassSchema lOCSchema = null;
        lEnum = lOutSchema.getObjectClasses();

        while (lEnum.hasMoreElements())
        {
          lOCSchema = (LDAPObjectClassSchema)lEnum.nextElement();
          lNode = DSMLLDAPConverter.objectclassToNode(lOCSchema, lDoc);
          lSchemaElem.appendChild(lNode);
        }
    
        /*********************************************************************/
        /* And also the attributes.                                          */
        /*********************************************************************/
        LDAPAttributeSchema lAttrSchema = null;
        lEnum = lOutSchema.getAttributes();
        
        while (lEnum.hasMoreElements())
        {
          lAttrSchema = (LDAPAttributeSchema)lEnum.nextElement();
          lNode = DSMLLDAPConverter.attrTypeToNode(lAttrSchema, lDoc);
          lSchemaElem.appendChild(lNode);
        }
      }
    }
    else
    {
      /***********************************************************************/
      /* If we aren't adding any sort of schema, we need to mark the         */
      /* document as incomplete.                                             */
      /***********************************************************************/
      (lDoc.getDocumentElement()).setAttribute("complete", "false");
    }

    /*************************************************************************/
    /* Return the Document                                                   */
    /*************************************************************************/
    return lDoc;

  } /* search() */
  
/** This method creates a schema which contains only those objectclasses
    and attributes referenced by the directory data.
    We perform this in LDAP rather than Document representation because
    it's almost certainly quicker to use native directory-oriented
    structures.

  @param xiEntries A Vector of LDAPEntries whose details need to go in the
    schema
  @param xiFullSchema An LDAPSchema which is a superset of that needed for the
    xiEntries

  @return An LDAPSchema object representing the mini schema, or null
  @return one could not be created.
  */
  private LDAPSchema createMiniSchema(Vector      xiEntries, 
                                      LDAPSchema  xiFullSchema)
  {
    /*************************************************************************/
    /* To create the mini schema we:                                         */
    /* - iterating over all search results, getting their objectclasses,     */
    /*   and storing as a HashSet (dupes removed automatically)              */
    /* - iterating over all objectclasses, getting their attributes, and     */
    /*   storing as a HashSet (again, dupes removed automatically)           */
    /* - copying all the found objectclasses and attributes from the         */
    /*   full schema to the mini schema.                                     */
    /*************************************************************************/    
    LDAPSchema            lMiniSchema = new LDAPSchema();  /* to be returned */

    Enumeration           lEnum       = null;
    Iterator              lIterator   = null; 

    HashSet               lAttrNames  = new HashSet();    /* attribute names */
    HashSet               lOCNames    = new HashSet();  /* objectclass names */

    LDAPEntry             lEntry      = null;
    LDAPAttribute         lAttr       = null;
    String                lAttrName   = null;
    String                lOCName     = null;
    LDAPObjectClassSchema lOCSchema   = null;

    /*************************************************************************/
    /* Iterate through the LDAPEntries returned from the search              */
    /*************************************************************************/
    lIterator = xiEntries.iterator();

    while(lIterator.hasNext())
    {
      lEntry = (LDAPEntry)lIterator.next();

      /***********************************************************************/
      /* For each one, get the objectclasses                                 */
      /***********************************************************************/
      lAttr = lEntry.getAttribute("objectclass"); 

      if (lAttr != null)
      {
        lEnum = lAttr.getStringValues();
      }
      else
      {
        /*********************************************************************/
        /* We should have made sure, above, that every entry has an          */
        /* objectclass attribute.                                            */
        /*********************************************************************/
        lEnum = null;
      }

      /***********************************************************************/
      /* Break down each set of objectclasses and add the names to our hash  */
      /***********************************************************************/
      if (lEnum != null) 
      {
        while (lEnum.hasMoreElements()) 
        {
          lOCName = (String)lEnum.nextElement();
          lOCNames.add(lOCName);       
        }
      }
    }

    
    /*************************************************************************/
    /* TEST_CODE                                                             */
    /*************************************************************************/
    // #trace(xiFullSchema.toString());
    
    /*************************************************************************/
    /* Iterate through the objectclasses of the search result entries        */
    /*************************************************************************/
    lIterator = lOCNames.iterator();

    while (lIterator.hasNext())
    {
      lEnum   = null;
      lOCName = (String)lIterator.next();

      /***********************************************************************/
      /* Get the LDAPObjectClassSchema and add it to the mini schema         */
      /***********************************************************************/
      lOCSchema = xiFullSchema.getObjectClass(lOCName);
      
      if (lOCSchema != null)
      {
        lMiniSchema.addObjectClass(lOCSchema);
      }
      else
      {
        System.err.println("Can't find objectclass \"" + lOCName + "\" in " + 
                           "directory schema.");
      }

      /***********************************************************************/
      /* Get all the attributes for each objectclass and add them to the     */
      /* hash.  Required attributes first.                                   */
      /***********************************************************************/
      if (lOCSchema != null)
      {
        lEnum = lOCSchema.getRequiredAttributes();
      }

      if (lEnum != null) 
      {
        while (lEnum.hasMoreElements()) 
        {
          lAttrName = (String)lEnum.nextElement();
          lAttrNames.add(lAttrName);              
        }
      }
      
      /***********************************************************************/
      /* And optional attributes, in the same way.                           */
      /***********************************************************************/
      if (lOCSchema != null)
      {
        lEnum = lOCSchema.getOptionalAttributes();
      }

      if (lEnum != null) 
      {
        while (lEnum.hasMoreElements()) 
        {
          lAttrName = (String)lEnum.nextElement();
          lAttrNames.add(lAttrName);              
        }
      }
    }

    
    /*************************************************************************/
    /* Finally, iterate through the attributes and add them to the mini      */
    /* schema as well                                                        */
    /*************************************************************************/
    lIterator = lAttrNames.iterator();

    while (lIterator.hasNext())
    {
      lAttrName = (String)lIterator.next();
      lMiniSchema.addAttribute(xiFullSchema.getAttribute(lAttrName));
    }

    /*************************************************************************/
    /* If we don't have any OCNames at all, then something has gone wrong,   */
    /* as we should have made sure when we called the search() that          */
    /* objectclasses were always returned.                                   */
    /*************************************************************************/
    if (lOCNames.isEmpty())
    {
      lMiniSchema = null;
    }

    return lMiniSchema;

  } /* createMiniSchema() */

/** Sets whether we output the full directory schema or a mini schema.

  @param xiOutputFullSchema The value to set it to
  */
  public void setOutputFullSchema(boolean xiOutputFullSchema)
  {
    m_outputFullSchema = xiOutputFullSchema;

  } /* setOutputFullSchema() */
  
/** Sets what sort of output is required.

  @param xiContents The type of output - one of <CODE>ALL</CODE>, <CODE>DATA
    </CODE> or <CODE>SCHEMA</CODE>.
  */
  public void setContents(int xiContents)
  {
    boolean lSafe;
    lSafe = (xiContents == SCHEMA || xiContents == DATA || xiContents == ALL);

    m_contents = xiContents;

  } /* setContents() */

}/* DSMLLDAPReader */
