/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.util.*;
import netscape.ldap.*;
import netscape.ldap.util.*;
import org.w3c.dom.*;
import org.dsmltools.util.DNLengthComparator;

/** 

  This class implements functionality to modify an LDAP directory based on
  the contents of a DSML document.
  */
public final class DSMLLDAPWriter extends DSMLLDAPConnection
{
/** 

  Flag to indicate when true, that:
  a) The add() method does not overwrite existing entries, and
  b) The delete() method makes sure of an exact match before deleting
  */
  private boolean m_careful = false;

/** This method adds the entries in a Document to the directory. <BR>
    <BR>
    If the careful flag is set, we do not overwrite entries which are
    already present in the directory (matching by DN.) If it is not set, we
    overwrite entries present in the directory totally.

  @param xiDoc Document containing entries to add
  */
  public void add(Document xiDoc)
  {
    /*************************************************************************/
    /* Local variables                                                       */
    /*                                                                       */
    /* All the variables in this first two sections are general reusables of */
    /* a specific type.                                                      */
    /*************************************************************************/
    LDAPObjectClassSchema lOCSchema   = null;
    LDAPAttributeSchema   lAttrSchema = null;
    
    Enumeration lEnum       = null;
    Iterator    lIterator   = null;
    Node        lNode       = null;
    Node        lOldNode    = null;
    NodeList    lNodes      = null;
    LDAPEntry   lEntry      = null;
    int         lLength     = 0;
    int         lCount      = 0;

    Vector      lEntries    = new Vector(); /* LDAPEntry objects to be added */
    Element     lDocElem    = null;               /* <dsml> Element of xiDoc */
    Node        lDSNode     = null;               /* <directory-schema> Node */
    LDAPSchema  lFullSchema = null;                /* the directory's schema */

    if (m_conn == null) 
    {
      System.err.println("Internal error: add() when not connected.");
      System.exit(1);
    }

    /*************************************************************************/
    /* Get full schema from directory                                        */
    /*************************************************************************/
    lFullSchema = new LDAPSchema();

    try
    {
      lFullSchema.fetchSchema(m_conn, "");          /* use root as search DN */
    }
    catch (LDAPException lExcept)
    {
      System.err.println("LDAPException while fetching schema: " + 
                         lExcept.toString());
    }

    /*************************************************************************/
    /* If we haven't been given any schema info by the directory, we need to */
    /* spot this and avoid trying the validate.  This can happen if you use  */
    /* Netscape's vanilla SDK which has a bug - it doesn't request all the   */
    /* schema info, and so only works with Netscape's directory.             */
    /*                                                                       */
    /* This test is not particularly efficient (calling getObjectClasses()), */
    /* but is the only way of spotting this problem.                         */
    /*************************************************************************/
    lEnum = lFullSchema.getObjectClasses();

    if (!lEnum.hasMoreElements())
    {
      System.err.println("Directory did not provide schema info; unable to " +
                         "assess compatibility of imported data.\n" +
                         "(Are you using an unpatched Netscape library?)");
    }
    else
    {
      /***********************************************************************/
      /* We need to validate the DSML data.  We do this by passing the       */
      /* directory's LDAPSchema as an "extra" parameter to validate().  This */
      /* stops validate() from calculating the schema from the Document we   */
      /* pass in (which is good, as it may not have one.)                    */
      /***********************************************************************/
      try
      {
        DSMLDocumentImpl.validate(xiDoc, lFullSchema);             
      }
      catch (DSMLRefException lExcept)
      {
        System.err.println("Unable to validate DSML data against directory " + 
                   "schema, as it contains references to an external schema.");
      }
    }

    /*************************************************************************/
    /* Store the entry nodes, as LDAPEntries, in a Vector                    */
    /*************************************************************************/
    lNodes = xiDoc.getElementsByTagName("entry");
    lLength = lNodes.getLength();

    for (lCount = 0; lCount < lLength; lCount++)
    {
      lNode  = lNodes.item(lCount);
      lEntry = DSMLLDAPConverter.nodeToEntry(lNode);
      lEntries.add(lEntry);
    }
        
    /*************************************************************************/
    /* Sort the Vector by length of DN in the LDAPEntries, smallest first.   */
    /*************************************************************************/
    DNLengthComparator lDNComparator = new DNLengthComparator(true);
    Collections.sort(lEntries, lDNComparator);

    /*************************************************************************/
    /* Iterate over the now-sorted list of LDAPEntries                       */
    /*************************************************************************/
    lIterator = lEntries.iterator();

    while (lIterator.hasNext())
    {
      lEntry = (LDAPEntry)lIterator.next();

      /***********************************************************************/
      /* Attempt the add                                                     */
      /***********************************************************************/
      try
      {
        m_conn.add(lEntry);
      }
      catch (LDAPException lExcept)
      {
        String lDN = lEntry.getDN();

        switch (lExcept.getLDAPResultCode())
        {
          /*******************************************************************/
          /* If the add fails with "entry already exists" and we are in      */
          /* overwrite mode, remove the entry and have another go.           */
          /*******************************************************************/
          case LDAPException.ENTRY_ALREADY_EXISTS:
            
            if (!m_careful)
            {
              updateEntry(lEntry);
            }
            else
            {
              System.err.println("Not overwriting entry \"" + lDN + "\".");
            }
            break;
          
          /*******************************************************************/
          /* Insufficient access rights error                                */
          /*******************************************************************/
          case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
            System.err.println("Can't add entry \"" + lDN + 
                            "\" as I don't have the necessary access rights.");
            break;

          /*******************************************************************/
          /* Otherwise, mention that we are not overwriting things.          */
          /*******************************************************************/
          default:
            System.err.println("LDAPException adding entry: " +
                               lExcept.toString());
            break;
        }
      }      
    }

  } /* add() */

/** This method deals with the case where we are adding an entry which
    already exists. It attemps to delete and then add the entry and, if
    that fails (probably because the entry is not a leaf) then it removes
    all its attributes and adds in the new ones.

  @param xiEntry The entry to update
  */
  private void updateEntry(LDAPEntry xiEntry)
  {
    String lDN = xiEntry.getDN();

    /*************************************************************************/
    /* First, attempt a simple delete then add - this will work for leaf     */
    /* nodes in the directory tree.                                          */
    /*************************************************************************/
    try 
    {
      m_conn.delete(lDN);
      m_conn.add(xiEntry);
    }
    catch (LDAPException lExcept)
    {
      if (lExcept.getLDAPResultCode() == LDAPException.NOT_ALLOWED_ON_NONLEAF)
      {
        /*********************************************************************/
        /* If that fails, we need to do a modify. This won't happen that     */
        /* often, so we don't create the necessary objects until now.        */
        /*********************************************************************/
        
        LDAPModificationSet lModSet   = new LDAPModificationSet();
                       /* LDAPModifications to change the current attributes */
        LDAPEntry          lDirEntry = null; /* the dir's version of xiEntry */
        LDAPAttribute      lAttr     = null;
        Enumeration        lAttrEnum = null;
        String             lName     = null;
  
        /*********************************************************************/
        /* Read the directory's version of this entry                        */
        /*********************************************************************/
        try 
        {
          lDirEntry = m_conn.read(lDN);
        }
        catch (LDAPException lExcept2)
        {
          /*******************************************************************/
          /* We know this DN exists because we just read it; this will       */
          /* therefore me some other sort of error.                          */
          /*******************************************************************/
          System.err.println("Suddenly unable to read from directory!");
          System.exit(1);
        }
  
        /*********************************************************************/
        /* Remove all the attributes the entry currently has by adding them  */
        /* to an LDAPModificationSet and (later) passing that to modify()    */
        /*********************************************************************/
        lAttrEnum = (lDirEntry.getAttributeSet()).getAttributes();
        
        while (lAttrEnum.hasMoreElements())
        {
          lAttr = (LDAPAttribute)lAttrEnum.nextElement();
          lName = lAttr.getName();
  
          /*******************************************************************/
          /* We aren't permitted to modify the objectclasses                 */
          /*******************************************************************/
          if (!(lName.equalsIgnoreCase("objectclass")))
          {
             lModSet.add(LDAPModification.DELETE, lAttr);
          }
        }
  
        /*********************************************************************/
        /* We need to add to the LDAPModificationSet all the new attributes  */
        /* (many of which may be the same as those just deleted, but this is */
        /* easier and faster than trying to do comparisions.)                */
        /*********************************************************************/
        lAttrEnum = (xiEntry.getAttributeSet()).getAttributes();
        
        while (lAttrEnum.hasMoreElements())
        {
          lAttr = (LDAPAttribute)lAttrEnum.nextElement();
          lName = lAttr.getName();
  
          /*******************************************************************/
          /* We aren't permitted to modify the objectclasses                 */
          /*******************************************************************/
          if (!(lName.equalsIgnoreCase("objectclass")))
          {
            lModSet.add(LDAPModification.ADD, lAttr);
          }
        }
  
        /*********************************************************************/
        /* Commit those changes                                              */
        /*********************************************************************/
        try
        {
          m_conn.modify(lDN, lModSet);
        }
        catch (LDAPException lExcept2)
        {
          switch (lExcept2.getLDAPResultCode())
          {
            /*****************************************************************/
            /* Insufficient access rights error                              */
            /*****************************************************************/
            case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
              System.err.println("Can't modify entry \"" + lDN + 
                            "\" as I don't have the necessary access rights.");
              break;
  
            /*****************************************************************/
            /* Otherwise, mention that we are not overwriting things.        */
            /*****************************************************************/
            default:
              System.err.println("LDAPException modifying entry: " + 
                                 lExcept2.toString());
              break;
          }
        }
      }
      else
      {
        System.err.println("LDAPException trying a del/add: " +
                                                           lExcept.toString());
      }
    }

  } /* updateEntry() */

/** This method removes the entries in the Document from the directory.
    For delete, we ignore the schema attached to the document.<BR>
    <BR>
    If the careful flag is not set, the deletion is done by matching DNs
    only. If the careful flag is set, the entries are compared, and only
    deleted if a perfect match is found. This second method takes
    significantly longer.

  @param xiDoc Document containing DSML data to be removed
  */
  public void delete(Document xiDoc)
  {
    /*************************************************************************/
    /* Local variables                                                       */
    /*************************************************************************/
    int       lCount  = 0;
    int       lLength = 0;
    
    String    lDN     = null;
    NodeList  lNodes  = null;
    Node      lNode   = null;

    LDAPEntry lDSMLEntry = null;     /* used for comparing DSML to directory */
    LDAPEntry lDirEntry  = null;     /* used for comparing DSML to directory */

    Vector    lEntries   = new Vector();    /* LDAPEntry objects to be added */

    Iterator  lIterator  = null;

    if (m_conn == null) 
    {
      System.err.println("Internal error: delete() when not connected.");
      System.exit(1);
    }

    /*************************************************************************/
    /* We have to sort the entries to make sure we delete leaf nodes first.  */
    /* We do this by sorting on length of DN.  However, this means that it's */
    /* easier to convert all the entries to LDAPEntries, even if we are only */
    /* deleting on the basis of DN.                                          */
    /*************************************************************************/
                                                                            
    /*************************************************************************/
    /* Store the entry nodes, as LDAPEntries, in a Vector                    */
    /*************************************************************************/
    lNodes  = xiDoc.getElementsByTagName("entry");
    lLength = lNodes.getLength();

    for (lCount = 0; lCount < lLength; lCount++)
    {
      lNode      = lNodes.item(lCount);
      lDSMLEntry = DSMLLDAPConverter.nodeToEntry(lNode);
      lEntries.add(lDSMLEntry);
    }
        
    /*************************************************************************/
    /* Sort the Vector by length of DN in the LDAPEntries, largest first.    */
    /*************************************************************************/
    DNLengthComparator lDNComparator = new DNLengthComparator(false);
    Collections.sort(lEntries, lDNComparator);

    /*************************************************************************/
    /* Iterate over the now-sorted list of LDAPEntries                       */
    /*************************************************************************/
    lIterator = lEntries.iterator();

    /*************************************************************************/
    /* What we do depends on whether we are being careful or not.            */
    /*************************************************************************/
    if (!m_careful)
    {
      /***********************************************************************/
      /* If we aren't, simply delete each entry on the basis of DN           */
      /***********************************************************************/
      
      while (lIterator.hasNext())
      {
        lDSMLEntry = (LDAPEntry)lIterator.next();
        lDN = lDSMLEntry.getDN();

        try 
        {
          m_conn.delete(lDN);
        }
        catch (LDAPException lExcept)
        {
          switch (lExcept.getLDAPResultCode())
          {
            /*****************************************************************/
            /* Non-existent object error                                     */
            /*****************************************************************/
            case LDAPException.NO_SUCH_OBJECT:
              System.err.println("Can't delete entry \"" + lDN + 
                                 "\" as it does not exist.");
              break;
          
            /*****************************************************************/
            /* Non-leaf deletion error                                       */
            /*****************************************************************/
            case LDAPException.NOT_ALLOWED_ON_NONLEAF:
              System.err.println("Can't delete entry \"" + lDN + 
                                 "\" as it's not a leaf node.");
              break;
            
            /*****************************************************************/
            /* Insufficient access rights error                              */
            /*****************************************************************/
            case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
              System.err.println("Can't delete entry \"" + lDN + 
                            "\" as I don't have the necessary access rights.");
              break;
            
            /*****************************************************************/
            /* Other errors                                                  */
            /*****************************************************************/
            default:
              System.err.println("Error deleting entry: " + 
                                 lExcept.toString());
              break;
          }
        }
      }
    }
    else
    {
      /***********************************************************************/
      /* If we are, we need to make sure the entries are exactly the same.   */
      /***********************************************************************/
      boolean lDeleteEntry;

      while (lIterator.hasNext())
      {
        lDSMLEntry = (LDAPEntry)lIterator.next();
        lDeleteEntry = false;

        /*********************************************************************/
        /* Try to get a matching directory entry                             */
        /*********************************************************************/
        lDirEntry = null;
        lDN       = lDSMLEntry.getDN();
        
        try
        {
          lDirEntry = m_conn.read(lDN);
        }
        catch (LDAPException lExcept)
        {
          switch (lExcept.getLDAPResultCode())
          {
            case LDAPException.NO_SUCH_OBJECT:
            {
              System.err.println("Can't delete entry \"" + lDN + 
                                 "\" as it does not exist.");
            }
            break;
          
            default:
              System.err.println("Error reading entry: " + lDN);
              break;
          }
        }

        if (lDirEntry != null)
        {
          /*******************************************************************/
          /* We've found an entry in the directory matching the one we want  */
          /* to delete. Now, we have to compare it with the one from the     */
          /* DSML to make sure all the fields are the same.                  */
          /*******************************************************************/
          
          /*******************************************************************/
          /* Compare the two entries, field for field                        */
          /*******************************************************************/        
          if (equalEntries(lDirEntry, lDSMLEntry))
          {
            lDeleteEntry = true;
          }
          else
          {
            /*****************************************************************/
            /* Otherwise, print an error message                             */
            /*****************************************************************/
            System.err.println("Entry \"" + lDN + "\" does not match entry " + 
                               "in directory - not deleting.");
          }
        }
        else
        {
          lDeleteEntry = true;
        }

        if (lDeleteEntry)
        {
          /*******************************************************************/
          /* If they match, delete this entry                                */
          /*******************************************************************/
          try
          {
            m_conn.delete(lDN);  
          }
          catch (LDAPException lExcept)
          {
            switch (lExcept.getLDAPResultCode())
            {
              /***************************************************************/
              /* Non-existent object error                                   */
              /***************************************************************/
              case LDAPException.NO_SUCH_OBJECT:
                System.err.println("Can't delete entry \"" + lDN + 
                                   "\" as it does not exist.");
                break;
            
              /***************************************************************/
              /* Non-leaf deletion error                                     */
              /***************************************************************/
              case LDAPException.NOT_ALLOWED_ON_NONLEAF:
                System.err.println("Can't delete entry \"" + lDN + 
                                   "\" as it's not a leaf node.");
                break;
              
              /***************************************************************/
              /* Insufficient access rights error                            */
              /***************************************************************/
              case LDAPException.INSUFFICIENT_ACCESS_RIGHTS:
                System.err.println("Can't delete entry \"" + lDN + 
                            "\" as I don't have the necessary access rights.");
                break;
              
              /***************************************************************/
              /* Other errors                                                */
              /***************************************************************/
              default:
                System.err.println("Error deleting entry: " + 
                                   lExcept.toString());
                break;
            }
          }
        }
      }
    }

  } /* delete() */

/** This function compares the LDAPEntry representations of two directory
    entries, returning true if they are equal, and false if they are not.

  @param xiEntry1 The first node to compare
  @param xiEntry2 The node to compare it to

  @return Whether the entries are equal or not
  */
  private boolean equalEntries(LDAPEntry xiEntry1, LDAPEntry xiEntry2)
  {
    /*************************************************************************/
    /* Local variables; numbers refer to relevance to xiEntry1 and xiEntry2  */
    /*************************************************************************/
    boolean          lMatch    = true;  /* whether the two LDAPEntries match */ 

    LDAPAttributeSet lAttrSet1 = null;
    LDAPAttributeSet lAttrSet2 = null;
    LDAPAttribute    lAttr1    = null;
    LDAPAttribute    lAttr2    = null;

    Enumeration      lEnum     = null;
    Enumeration      lAttrEnum = null;

    /*************************************************************************/
    /* First, compare DNs and sizes of attribute sets                        */
    /*************************************************************************/
    if ((xiEntry1.getDN()).equals(xiEntry2.getDN()))
    {
      
      lAttrSet1 = xiEntry1.getAttributeSet();
      lAttrSet2 = xiEntry2.getAttributeSet();

      if (lAttrSet1.size() != lAttrSet2.size())
      {
        lMatch = false;
      }
    }
    else
    {
      lMatch = false;
    }

    /*************************************************************************/
    /* If both of those are OK, start looking more closely. Because we know  */
    /* both have the same number of attributes, if we go through the         */
    /* attributes of one, matching with the other, and find matches for all, */
    /* we know the two are exactly the same.                                 */
    /*************************************************************************/
    if (lMatch == true)
    {
      lEnum = lAttrSet1.getAttributes();
      
      while (lMatch == true && lEnum.hasMoreElements())
      {
        lAttr1 = (LDAPAttribute)lEnum.nextElement();
        lAttr2 = lAttrSet2.getAttribute(lAttr1.getName());

        if (lAttr2 == null)
        {
          /*******************************************************************/
          /* We've found an attribute in Entry1 that's not in Entry2         */
          /*******************************************************************/
          lMatch = false;
        }
        else
        {
          /*******************************************************************/
          /* We have a match on attribute name. Now, compare the sizes and   */
          /* then the values                                                 */
          /*******************************************************************/
          if (!(lAttr1.size() == lAttr2.size()))
          {
            /*****************************************************************/
            /* Number of attributes match failure                            */
            /*****************************************************************/
            lMatch = false;
          }
          else
          {
            /*****************************************************************/
            /* Add all the values to a Set and then compare the Sets.        */
            /* Possibly not the most efficient way, but the simplest.        */
            /*****************************************************************/

            int lSize = lAttr1.size();
            HashSet Val1Set = new HashSet(lSize);
            HashSet Val2Set = new HashSet(lSize);

            lAttrEnum = lAttr1.getStringValues();
            while (lAttrEnum.hasMoreElements())
            {
              Val1Set.add(lAttrEnum.nextElement());
            }

            lAttrEnum = lAttr2.getStringValues();
            while (lAttrEnum.hasMoreElements())
            {
              Val2Set.add(lAttrEnum.nextElement());
            }

            /*****************************************************************/
            /* Compare the sets                                              */
            /*****************************************************************/
            if (!Val1Set.equals(Val2Set))
            {
              /***************************************************************/
              /* Individual attribute match failure                          */
              /***************************************************************/
              lMatch = false;
            }
            else
            {
            }
          }
        }
      }
    }

    return lMatch;

  } /* equalEntries() */

/** This method sets the Careful flag.  This flag has different meanings
    depending on whether we are adding or deleting (see above.)

  @param xiCareful The value to set the careful flag to
  */
  public void setCareful(boolean xiCareful)
  {
    m_careful = xiCareful;

  } /* setCareful() */
                                                                            
} /* DSMLLDAPWriter */
