/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import gnu.getopt.*;
import org.w3c.dom.Document;

/** 

  This class reads in a DSML file, converts it to a Document, validates it,
  and then writes it out again.
  <BR><BR>
  This program reads in DSML, either from a file or from stdin, validates
  it and then writes it out to a file or stdout.<BR>
  <BR><PRE>
  The command syntax is as follows. Bold indicates a mandatory option.<BR>
  DSML2LDAP infile outfile [options]<BR>
  Output options:<UL>
  <LI>--quick     -q Quick mode - do not validate DSML document
  <LI>--help      -? Prints a usage message.
  </UL>
  E.g.:
  DSML2LDAP in.dsml out.dsml
  </PRE>
  */
public class DSMLValidate extends DSMLFile
{
/** 

  The input file name
  */
  private static String  s_infileName  = null;
  
/** 

  The output file name
  */
  private static String  s_outfileName = null;
  
/** 

  True if we are validating, false if we are not.
  */
  private static boolean s_quick       = false;

/** This method drives the DSML2DSML process.

  @param xiArgs The command-line options
  */
  public static void main(String[] xiArgs)
  {
    Document lDoc = null;
  
    /*************************************************************************/
    /* Parse command line                                                    */
    /*************************************************************************/
    parseCommandLine(xiArgs);
    
    /*************************************************************************/
    /* Read in the DSML                                                      */
    /*************************************************************************/
    lDoc = readDSML(s_infileName);

    /*************************************************************************/
    /* Check for "complete" attribute; if not complete, don't validate.      */
    /*************************************************************************/
    if (((lDoc.getDocumentElement()).getAttribute("complete")).equals("false"))
    {
      s_quick = true;
      System.err.print("Unable to validate document");

      if (s_infileName != null)
      {
        System.err.print(" '" + s_infileName + "'");
      }

      System.err.println(" as it doesn't have a complete schema.");
    }

    /*************************************************************************/
    /* Validate if required                                                  */
    /*************************************************************************/
    if (!s_quick)
    {
      try
      {
        DSMLDocumentImpl.validate(lDoc);
      }
      catch (DSMLRefException lExcept)
      {
        System.err.print("Unable to validate document");
  
        if (s_infileName != null)
        {
          System.err.print(" '" + s_infileName + "'");
        }

        System.err.println(" as it contains references to external schema.");
      }
    }
  
    /*************************************************************************/
    /* Serialize the Document; it'll use stdout if outfile is null.          */
    /*************************************************************************/
    writeDSML(s_outfileName, lDoc);
  
  } /* main() */
 
/** This function creates a GetOpt object, passing in the argument string,
    and makes repeated calls to getopt() to set all or some of the
    command-line option member variables.

  @param xiArgs The argument string to parse
  */
  private static void parseCommandLine(String[] xiArgs)
  {
    int     lOpt           = 0;
    boolean lInfileDefined = false;

    Getopt lGetopt = new Getopt("DSML2DSML", 
                                xiArgs, 
                                "q",
                                null);

    /*************************************************************************/
    /* Keep parsing until we run out of options                              */
    /*************************************************************************/
    while ((lOpt = lGetopt.getopt()) != -1)
    {
      switch (lOpt)
      {
        /*********************************************************************/
        /* Turn off validation?                                              */
        /*********************************************************************/
        case 'q':
          s_quick = true;
          break;

        /*********************************************************************/
        /* Print usage message?                                              */
        /*                                                                   */
        /* This value can be requested on the command line, but is also      */
        /* returned when Getopt encounters an error.                         */
        /*********************************************************************/
        case '?':
          printUsageMessage();
          System.exit(0);
          break;
        
        /*********************************************************************/
        /* Default - invalid option                                          */
        /*********************************************************************/
        default:
          System.err.println("\nInvalid option: " + lOpt);
          System.exit(1);
          break;
      }
    }

    /*************************************************************************/
    /* Get non-option command line parameters.  There should be two of       */
    /* these.  Optind is left pointing at them.                              */
    /*************************************************************************/
    try
    {
      s_infileName = xiArgs[lGetopt.getOptind()];
      lInfileDefined = true;

      s_outfileName = xiArgs[lGetopt.getOptind() + 1];
    }
    catch (ArrayIndexOutOfBoundsException lExcept)
    {
      /***********************************************************************/
      /* If there are no non-option arguments, getopt sets Optind to one     */
      /* past the end of the array (not shockingly helpfully.) So, this      */
      /* exception means that 0 or 1 filenames were specified.  Zero is OK,  */
      /* as we can use stdin and stdout.                                     */
      /***********************************************************************/
      if (!lInfileDefined)
      {
        /*********************************************************************/
        /* No arguments given; we'll use stdin and stdout.                   */
        /*********************************************************************/
      }
      else
      {
        /*********************************************************************/
        /* One argument given; it's the same outfile as infile.              */
        /*********************************************************************/
        s_outfileName = s_infileName;
      }
    }

  } /* parseCommandLine() */
                                                                            
  /***************************************************************************/
  /* Method: printUsageMessage()                                             */
  /*                                                                         */
  /* This method prints a usage message - it should be called if the user    */
  /* incorrectly specifies one or more command-line parameters.              */
  /***************************************************************************/
  private static void printUsageMessage()
  {
    System.err.print(
    "\n" +
    "This program reads in DSML, either from a file or from stdin, validates\n" +
    "it and then writes it out to a file or stdout.\n\n" +
    "If you specify no filenames, stdin and stdout are used. If you specify " +
    "one filename, it is used as both the input and the output file. If you " +
    "specify two filenames, the first is the input file and the second is the " +
    "output file.\n\n" +
    "The command syntax is as follows (with commands in standard Unix format):\n" +
    "DSMLValidate <filename(s)> [options]\n" +
    "\n" +
    "Output options:\n" +
    "--quick      -q Quick mode - do not validate DSML documents\n" +
    "--help       -? Prints this usage message.\n" +
    "\n"
    );

  } /* printUsageMessage() */

} /* DSMLValidate */






