/*****************************************************************************/
/* The contents of this file are subject to the Mozilla Public License       */
/* Version 1.1 (the "License"); you may not use this file except in          */
/* compliance with the License.  You may obtain a copy of the License at     */
/* http://www.mozilla.org/MPL/                                               */
/*                                                                           */
/* Software distributed under the License is distributed on an "AS IS"       */
/* basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the  */
/* License for the specific language governing rights and limitations under  */
/* the License.                                                              */
/*                                                                           */
/* The Original Code is the DSML Tools.                                      */
/*                                                                           */
/* The Initial Developer of the Original Code is Data Connection Limited.    */
/* Portions created by Data Connection are Copyright (C) 2000 Data           */
/* Connection Limited.  All Rights Reserved.                                 */
/*                                                                           */
/* Contributor(s):                                                           */
/*                                                                           */
/* Alternatively, the contents of this file may be used under the terms of   */
/* the GNU General Public License Version 2 or later (the "GPL"), in which   */
/* case the provisions of the GPL are applicable instead of those above.  If */
/* you wish to allow use of your version of this file only under the terms   */
/* of the GPL and not to allow others to use your version of this file under */
/* the MPL, indicate your decision by deleting the provisions above and      */
/* replace them with the notice and other provisions required by the GPL.    */
/* If you do not delete the provisions above, a recipient may use your       */
/* version of this file under either the MPL or the GPL.                     */
/*                                                                           */
/* The Original Code was written to the Glory of God, for Data Connection,   */
/* by Gervase Markham <gerv@dsmltools.org>                                   */
/*****************************************************************************/
package org.dsmltools;

import java.io.*;
import java.util.*;
import gnu.getopt.*;
import netscape.ldap.*;

/** 

  This class exposes functionality to export a DSML file from a directory.
  <BR><BR>
  This program takes LDAP host and search parameters on the command line,
  connects to the given LDAP server and performs the search. It takes the
  results given and, according to options set, outputs them as a DSML file
  to stdout or a specified file.<BR>
  <BR><PRE>
  The command syntax is as follows. Bold indicates a mandatory option.<BR>
  LDAP2DSML outfile <B>-h ldap.example.com</B> [options]<BR>
  Connection options:<UL>
  <LI>--host        -h LDAP host.
  <LI>--port        -p Port on the server to connect to. Default - 389.
  <LI>--username    -u Username. Omit for anonymous login.
  <LI>--password    -w Password. Omit for anonymous login.
  </UL>
  Search options:<UL>
  <LI>--basedn      -b The base DN (distinguished name) for the search.
  Default - root.
  <LI>--filter      -f An LDAP filter string, e.g. "(objectclass=person)".
  Default - no filter.
  <LI>--attributes  -a A comma-separated list of attributes to return.
  Default - all user attributes.
  <LI>--scope       -s Defines the scope of the search. Followed by one of
  the following keywords:
  <UL> <LI>base - scope includes only the base DN
  <LI>one - scope includes base DN and entries one level below it
  <LI>sub - scope includes base DN and entries at all levels below it.
  Default.</UL>
  <LI>--deref       -d Defines how alias dereferencing is done during
  search. Followed by one of the following keywords:
  <UL> <LI>never - never dereference
  <LI>always - always dereference
  <LI>search - dereference during searching only. Default.
  <LI>find - dereference during finding of the base DN only.</UL>
  <LI>--maxresults  -z Maximum number of results to return. Default - all.
  (-z for compatibility with ldapsearch).
  </UL>
  Output options:<UL>
  <LI>--contains    -c Defines the contents of the DSML file. Followed by
  one of the following keywords:
  <UL> <LI>schema - just schema information
  <LI>data - just the data
  <LI>all - both schema and data. Default.</UL>
  <LI>--fullschema  -e Output the entire directory schema, not just the  
  bits referenced by the data.
  <LI>--quick       -q Quick mode - do not validate DSML documents
  <LI>--help        -? Prints a usage message.
  </UL>
  E.g.:
  LDAP2DSML -h ldap.ex.com -u "cn=Fred Bloggs" -w SecretWord -m 50 -s one
  -b "ou=Sales,o=Ace Ltd,c=us" -a "cn, email" &gt; out.dsml
  </PRE>
  */
public class LDAP2DSML extends DSMLFile
{
/** 

  Stores output file name
  */
  private static String   s_outfileName      = null;    /* default is stdout */

/** 

  Stores hostname to connect to
  */
  private static String   s_hostname         = null;            /* mandatory */
  
/** 

  Stores port to connect on
  */
  private static int      s_port             = LDAPv2.DEFAULT_PORT;
  
/** 

  Stores username to connect with (entry in directory)
  */
  private static String   s_username         = null;
  
/** 

  Stores password to connect with
  */
  private static String   s_password         = null;
  
/** 

  Stores the base DN for the LDAP search
  */
  private static String   s_baseDN           = new String();
  
/** 

  Stores the filter to use in the search
  */
  private static String   s_filter           = null;        
  
/** 

  Stores an array of attribute names - search will return only those
  attributes
  */
  private static String[] s_attributes       = null;
  
/** 

  Stores an array of attribute names - search will return only those
  attributes
  */
  private static boolean  s_attrsOnly        = false;
  
/** 

  Stores the maximum number of results that the search should return
  */
  private static int      s_maxResults       = 1000;
  
/** 

  Stores the scope of the search; one of LDAPv2.SCOPE_BASE,
  LDAPv2.SCOPE_SUB, and LDAPv2.SCOPE_ONE.
  */
  private static int      s_scope            = LDAPv2.SCOPE_SUB;
  
/** 

  Stores the style of dereferencing; one of LDAPv2.DEREF_NEVER,
  LDAPv2.DEREF_ALWAYS, LDAPv2.DEREF_SEARCHING and LDAPv2.DEREF_FINDING.
  */
  private static int      s_aliasDeref       = LDAPv2.DEREF_FINDING;
  
/** 

  Stores the form of output; one of DSMLLDAPReader.ALL,
  DSMLLDAPReader.SCHEMA and DSMLLDAPReader.DATA.
  */
  private static int      s_contents         = DSMLLDAPReader.ALL;
  
/** 

  Output the entire directory schema?
  */
  private static boolean  s_outputFullSchema = false;
  
/** 

  Avoid doing directory-style validation of the DSML?
  */
  private static boolean  s_quick            = false;

/** This method drives the LDAP2DSML process.

  @param xiArgs The command-line arguments
  */
  public static void main(String[] xiArgs)
  {
    /*************************************************************************/
    /* Parse command line                                                    */
    /*************************************************************************/
    parseCommandLine(xiArgs);

    /*************************************************************************/
    /* Create DSMLLDAPReader                                                 */
    /*************************************************************************/
    DSMLLDAPReader lReader = new DSMLLDAPReader();

    /*************************************************************************/
    /* Configure the DSMLLDAPReader.  It does its own error handling.        */
    /*************************************************************************/
    lReader.setOutputFullSchema(s_outputFullSchema);
    lReader.setContents(s_contents);

    /*************************************************************************/
    /* Call the search method                                                */
    /*************************************************************************/
    DSMLDocumentImpl lDoc = null;

    lReader.connect(s_hostname, s_port, s_username, s_password);

    LDAPSearchConstraints lCons = new LDAPSearchConstraints();
    lCons.setMaxResults(s_maxResults);
    lCons.setDereference(s_aliasDeref);
    lCons.setBatchSize(0);

    lDoc = lReader.search(s_baseDN,
                          s_scope, 
                          s_filter, 
                          s_attributes, 
                          s_attrsOnly,
                          lCons);

    lReader.disconnect();

    /*************************************************************************/
    /* Validate if requested on command-line                                 */
    /*************************************************************************/
    if (!s_quick && s_contents != DSMLLDAPReader.DATA)
    {
      System.err.print("Validating output document");
      if (s_outfileName != null)
      {
        System.err.print(" '" + s_outfileName + "'");
      }

      System.err.println(".");

      try
      {
        lDoc.validate();
      }
      catch (DSMLRefException lExcept)
      {
        System.err.println("Internal error: DSMLRefException on" + 
                           "generated document.");
      }
    }

    /*************************************************************************/
    /* Serialize the Document (if filename is null, it'll use stdout.)       */
    /*************************************************************************/
    writeDSML(s_outfileName, lDoc);

  } /* main() */

/** This function creates a Getopt object, passing in the argument string,
    and makes repeated calls to Getopt() to set all or some of the
    command-line option member variables.

  @param xiArgs The argument string to parse
  */
  private static void parseCommandLine(String[] xiArgs)
  {
    int lOpt = 0;

    LongOpt[] lLOs = new LongOpt[15];

    lLOs[0]  = new LongOpt("hostname",   LongOpt.REQUIRED_ARGUMENT, null, 'h');
    lLOs[1]  = new LongOpt("port",       LongOpt.REQUIRED_ARGUMENT, null, 'p');
    lLOs[2]  = new LongOpt("username",   LongOpt.REQUIRED_ARGUMENT, null, 'u');
    lLOs[3]  = new LongOpt("password",   LongOpt.REQUIRED_ARGUMENT, null, 'w');
    lLOs[4]  = new LongOpt("basedn",     LongOpt.REQUIRED_ARGUMENT, null, 'b');
    lLOs[5]  = new LongOpt("filter",     LongOpt.REQUIRED_ARGUMENT, null, 'f');
    lLOs[6]  = new LongOpt("attributes", LongOpt.REQUIRED_ARGUMENT, null, 'a');
    lLOs[7]  = new LongOpt("maxresults", LongOpt.REQUIRED_ARGUMENT, null, 'z');
    lLOs[8]  = new LongOpt("scope",      LongOpt.REQUIRED_ARGUMENT, null, 's');
    lLOs[9]  = new LongOpt("deref",      LongOpt.REQUIRED_ARGUMENT, null, 'd');
    lLOs[10] = new LongOpt("contains",   LongOpt.REQUIRED_ARGUMENT, null, 'c');
    lLOs[11] = new LongOpt("attrsonly",  LongOpt.NO_ARGUMENT, null, 'o');
    lLOs[12] = new LongOpt("fullschema", LongOpt.NO_ARGUMENT, null, 'e');
    lLOs[13] = new LongOpt("quick",      LongOpt.NO_ARGUMENT, null, 'q');
    lLOs[14] = new LongOpt("help",       LongOpt.NO_ARGUMENT, null, '?');

    Getopt lGetopt = new Getopt("LDAP2DSML", 
                                xiArgs, 
                                "h:p:u:w:b:f:a:z:s:d:c:oeq?",
                                lLOs);

    String lArg = null;

    while ((lOpt = lGetopt.getopt()) != -1)
    {
      switch (lOpt)
      {
        /*********************************************************************/
        /* Hostname                                                          */
        /*********************************************************************/
        case 'h':
          s_hostname = lGetopt.getOptarg();
          break;
      
        /*********************************************************************/
        /* Port                                                              */
        /*********************************************************************/
        case 'p':
          String lPort = lGetopt.getOptarg();
          try
          {
            s_port = Integer.parseInt(lPort);
          }
          catch (NumberFormatException lExcept)
          {
            System.err.println("LDAP2DSML: Invalid argument to -p: " + lPort +
                               ". Not a number.");
            printUsageMessage();
            System.exit(1);
          }
          break;

        /*********************************************************************/
        /* Username                                                          */
        /*********************************************************************/
        case 'u':
          s_username = lGetopt.getOptarg();
          break;

        /*********************************************************************/
        /* Password                                                          */
        /*********************************************************************/
        case 'w':
          s_password = lGetopt.getOptarg();
          break;

        /*********************************************************************/
        /* Base DN                                                           */
        /*********************************************************************/
        case 'b':
          s_baseDN = lGetopt.getOptarg();
          break;
        
        /*********************************************************************/
        /* Filter                                                            */
        /*********************************************************************/
        case 'f':
          s_filter = lGetopt.getOptarg();
          break;
          
        /*********************************************************************/
        /* Attributes; they need to be split up on ',' into an array.        */
        /*********************************************************************/
        case 'a':
          String          lAttrString = lGetopt.getOptarg();
          String          lAttr       = null;
          StringTokenizer lTokenizer  = new StringTokenizer(lAttrString, 
                                                            ",");
          Vector          lVect       = new Vector();

          while (lTokenizer.hasMoreElements())
          {
            lAttr = (lTokenizer.nextToken()).trim();
            lVect.add(lAttr);
          }

          s_attributes = new String[lVect.size()];
          lVect.copyInto(s_attributes);
          break;
        
        /*********************************************************************/
        /* Max results to return                                             */
        /*********************************************************************/
        case 'z':
          lArg = lGetopt.getOptarg();
          try
          {
            s_maxResults = Integer.parseInt(lArg);
          }
          catch (NumberFormatException lExcept)
          {
            System.err.println("LDAP2DSML: Invalid argument to -z: " + lArg +
                               ". Not a number.");
            printUsageMessage();
            System.exit(1);
          }
          break;
        
        /*********************************************************************/
        /* Scope                                                             */
        /*********************************************************************/
        case 's':
          lArg = lGetopt.getOptarg();
          if      (lArg.compareToIgnoreCase("base") == 0)
          {
            s_scope = LDAPv2.SCOPE_BASE;
          }
          else if (lArg.compareToIgnoreCase("one") == 0)
          {
            s_scope = LDAPv2.SCOPE_ONE;
          }
          else if (lArg.compareToIgnoreCase("sub") == 0)
          {
            s_scope = LDAPv2.SCOPE_SUB;
          }
          else
          {
            System.err.println("LDAP2DSML: Invalid argument to -s: " + lArg + 
                               ".");
            printUsageMessage();
            System.exit(1);
          }
          break;
        
        /*********************************************************************/
        /* How dereferencing is done                                         */
        /*********************************************************************/
        case 'd':
          lArg = lGetopt.getOptarg();
          if      (lArg.compareToIgnoreCase("never") == 0)
          {
            s_aliasDeref = LDAPv2.DEREF_NEVER;
          }
          else if (lArg.compareToIgnoreCase("always") == 0)
          {
            s_aliasDeref = LDAPv2.DEREF_ALWAYS;
          }
          else if (lArg.compareToIgnoreCase("search") == 0)
          {
            s_aliasDeref = LDAPv2.DEREF_SEARCHING;
          }
          else if (lArg.compareToIgnoreCase("find") == 0)
          {
            s_aliasDeref = LDAPv2.DEREF_FINDING;
          }
          else
          {
            System.err.println("LDAP2DSML: Invalid argument to -a: " + lArg + 
                               ".");
            printUsageMessage();
            System.exit(1);
          }
          break;
        
        /*********************************************************************/
        /* Contents of output file                                           */
        /*********************************************************************/
        case 'c':
          lArg = lGetopt.getOptarg();
          if      (lArg.compareToIgnoreCase("schema") == 0)
          {
            s_contents = DSMLLDAPReader.SCHEMA;
          }
          else if (lArg.compareToIgnoreCase("data") == 0)
          {
            s_contents = DSMLLDAPReader.DATA;
          }
          else if (lArg.compareToIgnoreCase("all") == 0)
          {
            s_contents = DSMLLDAPReader.ALL;
          }
          else
          {
            System.err.println("LDAP2DSML: Invalid argument to -c: " + lArg + 
                               ".");
            printUsageMessage();
            System.exit(1);
          }
          break;
        
        /*********************************************************************/
        /* Output full schema?                                               */
        /*********************************************************************/
        case 'e':
          s_outputFullSchema = true;
          break;
        
        /*********************************************************************/
        /* Attributes only?                                                  */
        /*********************************************************************/
        case 'o':
          s_attrsOnly = true;
          break;
        
        /*********************************************************************/
        /* Turn off validation?                                              */
        /*********************************************************************/
        case 'q':
          s_quick = true;
          break;

        /*********************************************************************/
        /* Print usage message?                                              */
        /*                                                                   */
        /* This value can be requested on the command line, but is also      */
        /* returned when Getopt encounters an error.                         */
        /*********************************************************************/
        case '?':
          printUsageMessage();
          System.exit(0);
          break;
        
        /*********************************************************************/
        /* Default - invalid option                                          */
        /*********************************************************************/
        default:
          System.err.println("LDAP2DSML: Invalid option: " + lOpt);
          printUsageMessage();
          System.exit(1);
          break;
      }
    }

    /*************************************************************************/
    /* Get non-option command line parameter (outfilename).  There should    */
    /* only be one of these.  Optind is left pointing at it.                 */
    /*************************************************************************/
    int lIndex = lGetopt.getOptind();

    try
    {
      s_outfileName = xiArgs[lIndex++];
    }
    catch (ArrayIndexOutOfBoundsException lExcept)
    {
      /***********************************************************************/
      /* If there are no non-option arguments, getopt sets Optind to one     */
      /* past the end of the array (not shockingly helpfully.) So, this      */
      /* exception means that no outfileName was specified; we'll use        */
      /* stdout.                                                             */
      /***********************************************************************/
      s_outfileName = null;
    }

    /*************************************************************************/
    /* If we haven't fallen off the end of xiArgs yet, there are too many    */
    /* command-line parameters.                                              */
    /*************************************************************************/
    if (lIndex < xiArgs.length)
    {
      System.err.println("LDAP2DSML: Too many filenames specified.");
      printUsageMessage();
      System.exit(1);
    }

    /*************************************************************************/
    /* Check for mandatory attributes                                        */
    /*************************************************************************/
    if (s_hostname == null)
    {
      System.err.println("LDAP2DSML: You must supply a hostname.");
      printUsageMessage();
      System.exit(1);
    }    

  } /* parseCommandLine() */

  /***************************************************************************/
  /* Method: printUsageMessage()                                             */
  /*                                                                         */
  /* This method prints a usage message - it should be called if the user    */
  /* incorrectly specifies one or more command-line parameters.              */
  /***************************************************************************/
  private static void printUsageMessage()
  {
    System.err.print(
    "\n" +
    "This command reads data from an LDAP directory and writes it as a DSML file.\n" +
    "\n" +
    "The command syntax is as follows (with commands in standard Unix format):\n" +
    "LDAP2DSML <outfile> -h ldap.example.com [options]\n" +
    "\n" +
    "Connection options:\n" +
    "--host       -h LDAP host. Mandatory.\n" +
    "--port       -p Port on the server to connect to. Default - 389.\n" +
    "--username   -u Username. Omit for anonymous login.\n" +
    "--password   -w Password. Omit for anonymous login.\n" +
    "Search options:\n" +
    "--basedn     -b The base DN (distinguished name) for the search.\n" + 
    "                Default - root.\n" +
    "--filter     -f An LDAP filter string, e.g. \"(objectclass=person)\".\n" + 
    "                Default - no filter.\n" +
    "--attributes -a A comma-separated list of attributes to return.\n" + 
    "                Default - all user attributes.\n" +
    "--attrsonly  -o Whether you want 'attributes only' output. Default - false.\n" +
    "--maxresults -z Maximum number of results to return. Default - all.\n" +
    "--scope      -s Defines the scope of the search.\n" + 
    "                Followed by one of the following keywords:\n" +
    "                base - only the base DN\n" +
    "                one - base DN and entries one level below it\n" +
    "                sub - base DN and entries at all levels below. Default.\n" + 
    "--aliasderef -a Defines how alias dereferencing is done during search. \n" +
    "                Followed by one of the following keywords:\n" +
    "                never - never dereference\n" +
    "                always - always dereference\n" +
    "                search - dereference during searching only. Default.\n" +
    "                find - dereference during finding of the base DN only.\n" +
    "Output options:\n" +
    "--contains   -c Defines the contents of the DSML file.\n" + 
    "                Followed by one of the following keywords:\n" +
    "                schema - just schema information\n" +
    "                data - just the data\n" +
    "                all - both schema and data. Default.\n" +
    "--fullschema -e Output the entire directory schema, not just the bits\n" + 
    "                referenced by the data.\n" +
    "--quick      -q Quick mode - do not validate DSML documents\n" +
    "--help       -? Prints this usage message.\n" +
    "\n"
    );

  } /* printUsageMessage() */

  /***************************************************************************/
  /* TEST_CODE                                                               */
  /***************************************************************************/
  private static void printMemberVariables()
  {      
    System.err.print("s_outfileName: ");
    System.err.println(s_outfileName);
    
    System.err.print("s_hostname: ");
    System.err.println(s_hostname);
    System.err.print("s_port: ");
    System.err.println(s_port);
    System.err.print("s_username: ");
    System.err.println(s_username);
    System.err.print("s_password: ");
    System.err.println(s_password);

    System.err.print("s_baseDN: ");
    System.err.println(s_baseDN);
    System.err.print("s_filter: ");
    System.err.println(s_filter);
    System.err.println("s_attributes: ");

    if (s_attributes != null)
    {
      Iterator lIterator = (Arrays.asList(s_attributes)).iterator();

      while (lIterator.hasNext())
      {
        System.err.println((String)lIterator.next());
      }
    }

    System.err.print("s_maxResults: ");
    System.err.println(s_maxResults);
    System.err.print("s_scope: ");
    System.err.println(s_scope);
    System.err.print("s_aliasDeref: ");
    System.err.println(s_aliasDeref);

    System.err.print("s_contents: ");
    System.err.println(s_contents);
    System.err.print("s_outputFullSchema: ");
    System.err.println(s_outputFullSchema);
    System.err.print("s_quick: ");
    System.err.println(s_quick);
  }

} /* LDAP2DSML */

